try:
    import zlib
except ImportError:
    try:
        import uzlib as zlib
    except ImportError:
        print("SKIP")
        raise SystemExit

PATTERNS = [
    # Packed results produced by CPy's zlib.compress()
    (b"0", b"x\x9c3\x00\x00\x001\x001"),
    (b"a", b"x\x9cK\x04\x00\x00b\x00b"),
    (b"0" * 100, b"x\x9c30\xa0=\x00\x00\xb3q\x12\xc1"),
    (
        bytes(range(64)),
        b"x\x9cc`dbfaec\xe7\xe0\xe4\xe2\xe6\xe1\xe5\xe3\x17\x10\x14\x12\x16\x11\x15\x13\x97\x90\x94\x92\x96\x91\x95\x93WPTRVQUS\xd7\xd0\xd4\xd2\xd6\xd1\xd5\xd370426153\xb7\xb0\xb4\xb2\xb6\xb1\xb5\xb3\x07\x00\xaa\xe0\x07\xe1",
    ),
    (b"hello", b"x\x01\x01\x05\x00\xfa\xffhello\x06,\x02\x15"),  # compression level 0
    # adaptive/dynamic huffman tree
    (
        b"13371813150|13764518736|12345678901",
        b"x\x9c\x05\xc1\x81\x01\x000\x04\x04\xb1\x95\\\x1f\xcfn\x86o\x82d\x06Qq\xc8\x9d\xc5X}<e\xb5g\x83\x0f\x89X\x07\xab",
    ),
    # dynamic Huffman tree with "case 17" (repeat code for 3-10 times)
    (
        b">I}\x00\x951D>I}\x00\x951D>I}\x00\x951D>I}\x00\x951D",
        b"x\x9c\x05\xc11\x01\x00\x00\x00\x010\x95\x14py\x84\x12C_\x9bR\x8cV\x8a\xd1J1Z)F\x1fw`\x089",
    ),
]

for unpacked, packed in PATTERNS:
    assert zlib.decompress(packed) == unpacked
    print(unpacked)


# Raw DEFLATE bitstream
v = b"\xcbH\xcd\xc9\xc9\x07\x00"
exp = b"hello"
out = zlib.decompress(v, -15)
assert out == exp
print(exp)
# Even when you ask CPython zlib.compress to produce Raw DEFLATE stream,
# it returns it with adler2 and oriignal size appended, as if it was a
# zlib stream. Make sure there're no random issues decompressing such.
v = b"\xcbH\xcd\xc9\xc9\x07\x00\x86\xa6\x106\x05\x00\x00\x00"
out = zlib.decompress(v, -15)
assert out == exp

# this should error
try:
    zlib.decompress(b"abc")
except Exception:
    print("Exception")

# invalid block type
try:
    zlib.decompress(b"\x07", -15)  # final-block, block-type=3 (invalid)
except Exception as er:
    print("Exception")



# test _thread.get_ident() function
#
# SPDX-FileCopyrightText: Copyright (c) 2016 Damien P. George on behalf of Pycom Ltd
#
# SPDX-License-Identifier: MIT

import _thread


def thread_entry():
    tid = _thread.get_ident()
    print("thread", type(tid) == int, tid != 0, tid != tid_main)
    global finished
    finished = True


tid_main = _thread.get_ident()
print("main", type(tid_main) == int, tid_main != 0)

finished = False
_thread.start_new_thread(thread_entry, ())

while not finished:
    pass
print("done")



"""
categories: Core,import
description: Failed to load modules are still registered as loaded
cause: To make module handling more efficient, it's not wrapped with exception handling.
workaround: Test modules before production use; during development, use ``del sys.modules["name"]``, or just soft or hard reset the board.
"""
import sys

try:
    from modules import foo
except NameError as e:
    print(e)
try:
    from modules import foo

    print("Should not get here")
except NameError as e:
    print(e)



# outer generator ignores a thrown GeneratorExit (this is allowed)

def gen():
    try:
        yield 123
    except GeneratorExit:
        print('GeneratorExit')

def gen2():
    try:
        yield from gen()
    except GeneratorExit:
        print('GeneratorExit outer')
    yield 789

# thrown a class
g = gen2()
print(next(g))
print(g.throw(GeneratorExit))

# thrown an instance
g = gen2()
print(next(g))
print(g.throw(GeneratorExit()))



# test struct package with floats
try:
    try:
        import ustruct as struct
    except:
        import struct
except ImportError:
    print("SKIP")
    raise SystemExit

i = 1.0 + 1 / 2
# TODO: it looks like '=' format modifier is not yet supported
# for fmt in ('f', 'd', '>f', '>d', '<f', '<d', '=f', '=d'):
for fmt in ("f", "d", ">f", ">d", "<f", "<d"):
    x = struct.pack(fmt, i)
    v = struct.unpack(fmt, x)[0]
    print("%2s: %.17f - %s" % (fmt, v, (i == v) and "passed" or "failed"))



# Test for utimeq module which implements task queue with support for
# wraparound time (utime.ticks_ms() style).
try:
    from utime import ticks_add, ticks_diff
    from utimeq import utimeq
except ImportError:
    print("SKIP")
    raise SystemExit

DEBUG = 0

MAX = ticks_add(0, -1)
MODULO_HALF = MAX // 2 + 1

if DEBUG:

    def dprint(*v):
        print(*v)


else:

    def dprint(*v):
        pass


# Try not to crash on invalid data
h = utimeq(10)
try:
    h.push(1)
    assert False
except TypeError:
    pass

try:
    h.pop(1)
    assert False
except IndexError:
    pass

# unsupported unary op
try:
    ~h
    assert False
except TypeError:
    pass

# pushing on full queue
h = utimeq(1)
h.push(1, 0, 0)
try:
    h.push(2, 0, 0)
    assert False
except IndexError:
    pass

# popping into invalid type
try:
    h.pop([])
    assert False
except TypeError:
    pass

# length
assert len(h) == 1

# peektime
assert h.peektime() == 1

# peektime with empty queue
try:
    utimeq(1).peektime()
    assert False
except IndexError:
    pass


def pop_all(h):
    l = []
    while h:
        item = [0, 0, 0]
        h.pop(item)
        # print("!", item)
        l.append(tuple(item))
    dprint(l)
    return l


def add(h, v):
    h.push(v, 0, 0)
    dprint("-----")
    # h.dump()
    dprint("-----")


h = utimeq(10)
add(h, 0)
add(h, MAX)
add(h, MAX - 1)
add(h, 101)
add(h, 100)
add(h, MAX - 2)
dprint(h)
l = pop_all(h)
for i in range(len(l) - 1):
    diff = ticks_diff(l[i + 1][0], l[i][0])
    assert diff > 0


def edge_case(edge, offset):
    h = utimeq(10)
    add(h, ticks_add(0, offset))
    add(h, ticks_add(edge, offset))
    dprint(h)
    l = pop_all(h)
    diff = ticks_diff(l[1][0], l[0][0])
    dprint(diff, diff > 0)
    return diff


dprint("===")
diff = edge_case(MODULO_HALF - 1, 0)
assert diff == MODULO_HALF - 1
assert edge_case(MODULO_HALF - 1, 100) == diff
assert edge_case(MODULO_HALF - 1, -100) == diff

# We expect diff to be always positive, per the definition of heappop() which should return
# the smallest value.
# This is the edge case where this invariant breaks, due to assymetry of two's-complement
# range - there's one more negative integer than positive, so heappushing values like below
# will then make ticks_diff() return the minimum negative value. We could make heappop
# return them in a different order, but ticks_diff() result would be the same. Conclusion:
# never add to a heap values where (a - b) == MODULO_HALF (and which are >= MODULO_HALF
# ticks apart in real time of course).
dprint("===")
diff = edge_case(MODULO_HALF, 0)
assert diff == -MODULO_HALF
assert edge_case(MODULO_HALF, 100) == diff
assert edge_case(MODULO_HALF, -100) == diff

dprint("===")
diff = edge_case(MODULO_HALF + 1, 0)
assert diff == MODULO_HALF - 1
assert edge_case(MODULO_HALF + 1, 100) == diff
assert edge_case(MODULO_HALF + 1, -100) == diff

print("OK")



# test compile builtin

try:
    compile
except NameError:
    print("SKIP")
    raise SystemExit

def test():
    global x

    c = compile("print(x)", "file", "exec")

    try:
        exec(c)
    except NameError:
        print("NameError")

    # global variable for compiled code to access
    x = 1

    exec(c)

    exec(c, {"x":2})
    exec(c, {}, {"x":3})

    # single/eval mode
    exec(compile("if 1: 10 + 1\n", "file", "single"))
    exec(compile("print(10 + 2)", "file", "single"))
    print(eval(compile("10 + 3", "file", "eval")))

    # bad mode
    try:
        compile('1', 'file', '')
    except ValueError:
        print("ValueError")

    # exception within compiled code
    try:
        exec(compile('noexist', 'file', 'exec'))
    except NameError:
        print("NameError")
    print(x) # check 'x' still exists as a global

test()



# Reraising last exception with raise w/o args

def f():
    try:
        raise ValueError("val", 3)
    except:
        raise

try:
    f()
except ValueError as e:
    print(repr(e))


# Can reraise only in except block
try:
    raise
except RuntimeError:
    print("RuntimeError")



# This tests that recursion with iternext doesn't lead to segfault.
try:
    enumerate
    filter
    map
    max
    zip
except:
    print("SKIP")
    raise SystemExit

# We need to pick an N that is large enough to hit the recursion
# limit, but not too large that we run out of heap memory.
try:
    # large stack/heap, eg unix
    [0] * 80000
    N = 5000
except:
    try:
        # medium, eg pyboard
        [0] * 10000
        N = 1000
    except:
        # small, eg esp8266
        N = 100

try:
    x = (1, 2)
    for i in range(N):
        x = enumerate(x)
    tuple(x)
except RuntimeError:
    print("RuntimeError")

try:
    x = (1, 2)
    for i in range(N):
        x = filter(None, x)
    tuple(x)
except RuntimeError:
    print("RuntimeError")

try:
    x = (1, 2)
    for i in range(N):
        x = map(max, x, ())
    tuple(x)
except RuntimeError:
    print("RuntimeError")

try:
    x = (1, 2)
    for i in range(N):
        x = zip(x)
    tuple(x)
except RuntimeError:
    print("RuntimeError")



try:
    extra_coverage
except NameError:
    print("SKIP")
    raise SystemExit

import uerrno
import uio

data = extra_coverage()

# test hashing of str/bytes that have an invalid hash
print(data[0], data[1])
print(hash(data[0]))
print(hash(data[1]))
print(hash(bytes(data[0], "utf8")))
print(hash(str(data[1], "utf8")))

# test streams
stream = data[2]  # has set_error and set_buf. Write always returns error
stream.set_error(uerrno.EAGAIN)  # non-blocking error
print(stream.read())  # read all encounters non-blocking error
print(stream.read(1))  # read 1 byte encounters non-blocking error
print(stream.readline())  # readline encounters non-blocking error
print(stream.readinto(bytearray(10)))  # readinto encounters non-blocking error
print(stream.write(b"1"))  # write encounters non-blocking error
print(stream.write1(b"1"))  # write1 encounters non-blocking error
stream.set_buf(b"123")
print(stream.read(4))  # read encounters non-blocking error after successful reads
stream.set_buf(b"123")
print(stream.read1(4))  # read1 encounters non-blocking error after successful reads
stream.set_buf(b"123")
print(stream.readline(4))  # readline encounters non-blocking error after successful reads
try:
    print(stream.ioctl(0, 0))  # ioctl encounters non-blocking error; raises OSError
except OSError:
    print("OSError")
stream.set_error(0)
print(stream.ioctl(0, bytearray(10)))  # successful ioctl call

stream2 = data[3]  # is textio
print(stream2.read(1))  # read 1 byte encounters non-blocking error with textio stream

# test BufferedWriter with stream errors
stream.set_error(uerrno.EAGAIN)
buf = uio.BufferedWriter(stream, 8)
print(buf.write(bytearray(16)))

# function defined in C++ code
print("cpp", extra_cpp_coverage())

# test user C module
import cexample

print(cexample.add_ints(3, 2))

# test user C module mixed with C++ code
import cppexample

print(cppexample.cppfunc(1, 2))

# test basic import of frozen scripts
import frzstr1

print(frzstr1.__file__)
import frzmpy1

print(frzmpy1.__file__)

# test import of frozen packages with __init__.py
import frzstr_pkg1

print(frzstr_pkg1.__file__, frzstr_pkg1.x)
import frzmpy_pkg1

print(frzmpy_pkg1.__file__, frzmpy_pkg1.x)

# test import of frozen packages without __init__.py
from frzstr_pkg2.mod import Foo

print(Foo.x)
from frzmpy_pkg2.mod import Foo

print(Foo.x)

# test raising exception in frozen script
try:
    import frzmpy2
except ZeroDivisionError:
    print("ZeroDivisionError")

# test loading a resource from a frozen string
import uio

buf = uio.resource_stream("frzstr_pkg2", "mod.py")
print(buf.read(21))

# test for MP_QSTR_NULL regression
from frzqstr import returns_NULL

print(returns_NULL())



try:
    import ure as re
except ImportError:
    try:
        import re
    except ImportError:
        print("SKIP")
        raise SystemExit

r = re.compile(" ")
s = r.split("a b c foobar")
print(s)

r = re.compile(" +")
s = r.split("a b    c   foobar")
print(s)

r = re.compile(" +")
s = r.split("a b    c   foobar", 1)
print(s)

r = re.compile(" +")
s = r.split("a b    c   foobar", 2)
print(s)

r = re.compile("[a-f]+")
s = r.split("0a3b9")
print(s)

# bytes objects
r = re.compile(b"x")
s = r.split(b"fooxbar")
print(s)



# Test asyncio.wait_for, with forwarding cancellation

try:
    import uasyncio as asyncio
except ImportError:
    try:
        import asyncio
    except ImportError:
        print("SKIP")
        raise SystemExit


async def awaiting(t, return_if_fail):
    try:
        print("awaiting started")
        await asyncio.sleep(t)
    except asyncio.CancelledError as er:
        # CPython wait_for raises CancelledError inside task but TimeoutError in wait_for
        print("awaiting canceled")
        if return_if_fail:
            return False  # return has no effect if Cancelled
        else:
            raise er
    except Exception as er:
        print("caught exception", er)
        raise er


async def test_cancellation_forwarded(catch, catch_inside):
    print("----------")

    async def wait():
        try:
            await asyncio.wait_for(awaiting(2, catch_inside), 1)
        except asyncio.TimeoutError as er:
            print("Got timeout error")
            raise er
        except asyncio.CancelledError as er:
            print("Got canceled")
            if not catch:
                raise er

    async def cancel(t):
        print("cancel started")
        await asyncio.sleep(0.01)
        print("cancel wait()")
        t.cancel()

    t = asyncio.create_task(wait())
    k = asyncio.create_task(cancel(t))
    try:
        await t
    except asyncio.CancelledError:
        print("waiting got cancelled")


asyncio.run(test_cancellation_forwarded(False, False))
asyncio.run(test_cancellation_forwarded(False, True))
asyncio.run(test_cancellation_forwarded(True, True))
asyncio.run(test_cancellation_forwarded(True, False))



# With MICROPY_CPYTHON_COMPAT, the "return" statement can only appear in a
# function.
# Otherwise (in minimal builds), it ends execution of a module/class.

try:
    exec
except NameError:
    print("SKIP")
    raise SystemExit

try:
    exec('return; print("this should not be executed.")')
    # if we get here then MICROPY_CPYTHON_COMPAT is disabled and test
    # should be skipped.
    print("SKIP")
    raise SystemExit
except SyntaxError:
    print('SyntaxError')



# A contains everything
class A:
    def __contains__(self, key):
        return True

a = A()
print(True in a)
print(1 in a)
print(() in a)

# B contains given things
class B:
    def __init__(self, items):
        self.items = items
    def __contains__(self, key):
        return key in self.items

b = B([])
print(1 in b)
b = B([1, 2])
print(1 in b)
print(2 in b)
print(3 in b)



@micropython.asm_thumb  # r0 = (int)(r0*r1/r2)
def muldiv(r0, r1, r2):
    vmov(s0, r0)
    vcvt_f32_s32(s0, s0)
    vmov(s1, r1)
    vcvt_f32_s32(s1, s1)
    vmov(s2, r2)
    vcvt_f32_s32(s2, s2)
    vmul(s7, s0, s1)
    vdiv(s8, s7, s2)
    vcvt_s32_f32(s31, s8)
    vmov(r0, s31)


print(muldiv(100, 10, 50))



"""
categories: Types,int
description: No int conversion for int-derived types available
cause: Unknown
workaround: Avoid subclassing builtin types unless really needed. Prefer https://en.wikipedia.org/wiki/Composition_over_inheritance .
"""


class A(int):
    __add__ = lambda self, other: A(int(self) + other)


a = A(42)
print(a + a)



# test ptr8 type


@micropython.viper
def set(dest: ptr8, val: int):
    dest[0] = val


@micropython.viper
def set1(dest: ptr8, val: int):
    dest[1] = val


@micropython.viper
def memset(dest: ptr8, val: int, n: int):
    for i in range(n):
        dest[i] = val


@micropython.viper
def memset2(dest_in, val: int):
    dest = ptr8(dest_in)
    n = int(len(dest_in))
    for i in range(n):
        dest[i] = val


b = bytearray(4)
print(b)

set(b, 41)
print(b)

set1(b, 42)
print(b)

memset(b, 43, len(b))
print(b)

memset2(b, 44)
print(b)



try:
    bytes.count
except AttributeError:
    print("SKIP")
    raise SystemExit

print(b"".count(b""))
print(b"".count(b"a"))
print(b"a".count(b""))
print(b"a".count(b"a"))
print(b"a".count(b"b"))
print(b"b".count(b"a"))

print(b"aaa".count(b""))
print(b"aaa".count(b"a"))
print(b"aaa".count(b"aa"))
print(b"aaa".count(b"aaa"))
print(b"aaa".count(b"aaaa"))

print(b"aaaa".count(b""))
print(b"aaaa".count(b"a"))
print(b"aaaa".count(b"aa"))
print(b"aaaa".count(b"aaa"))
print(b"aaaa".count(b"aaaa"))
print(b"aaaa".count(b"aaaaa"))

print(b"aaa".count(b"", 1))
print(b"aaa".count(b"", 2))
print(b"aaa".count(b"", 3))

print(b"aaa".count(b"", 1, 2))

print(b"asdfasdfaaa".count(b"asdf", -100))
print(b"asdfasdfaaa".count(b"asdf", -8))
print(b"asdf".count(b's', True))
print(b"asdf".count(b'a', True))
print(b"asdf".count(b'a', False))
print(b"asdf".count(b'a', 1 == 2))
print(b"hello world".count(b'l'))
print(b"hello world".count(b'l', 5))
print(b"hello world".count(b'l', 3))
print(b"hello world".count(b'z', 3, 6))
print(b"aaaa".count(b'a'))
print(b"aaaa".count(b'a', 0, 3))
print(b"aaaa".count(b'a', 0, 4))
print(b"aaaa".count(b'a', 0, 5))
print(b"aaaa".count(b'a', 1, 5))
print(b"aaaa".count(b'a', -1, 5))
print(b"abbabba".count(b"abba"))

def t():
    return True

print(b"0000".count(b'0', t()))



class mylist(list):
    pass

a = mylist([1, 2, 5])
# Test setting instance attr
a.attr = "something"
# Test base type __str__
print(a)
# Test getting instance attr
print(a.attr)
# Test base type ->subscr
print(a[-1])
a[0] = -1
print(a)
# Test another base type unary op
print(len(a))

# Test binary op of base type, with 2nd arg being raw base type
print(a + [20, 30, 40])
# Test binary op of base type, with 2nd arg being same class as 1st arg
# TODO: Faults
#print(a + a)

def foo():
    print("hello from foo")

try:
    class myfunc(type(foo)):
        pass
except TypeError:
    print("TypeError")

# multiple bases with layout conflict
try:
    class A(type, tuple):
        None
except TypeError:
    print('TypeError')



# test errors with import *

# 'import *' is not allowed in function scope
try:
    exec("def foo(): from x import *")
except SyntaxError as er:
    print("function", "SyntaxError")

# 'import *' is not allowed in class scope
try:
    exec("class C: from x import *")
except SyntaxError as er:
    print("class", "SyntaxError")



try:
    import framebuf
except ImportError:
    print("SKIP")
    raise SystemExit


def printbuf():
    print("--8<--")
    for y in range(h):
        for x in range(w):
            print("%u" % ((buf[(x + y * w) // 4] >> ((x & 3) << 1)) & 3), end="")
        print()
    print("-->8--")


w = 8
h = 5
buf = bytearray(w * h // 4)
fbuf = framebuf.FrameBuffer(buf, w, h, framebuf.GS2_HMSB)

# fill
fbuf.fill(3)
printbuf()
fbuf.fill(0)
printbuf()

# put pixel
fbuf.pixel(0, 0, 1)
fbuf.pixel(3, 0, 2)
fbuf.pixel(0, 4, 3)
fbuf.pixel(3, 4, 2)
printbuf()

# get pixel
print(fbuf.pixel(0, 4), fbuf.pixel(1, 1))

# scroll
fbuf.fill(0)
fbuf.pixel(2, 2, 3)
printbuf()
fbuf.scroll(0, 1)
printbuf()
fbuf.scroll(1, 0)
printbuf()
fbuf.scroll(-1, -2)
printbuf()

w2 = 2
h2 = 3
buf2 = bytearray(w2 * h2 // 4)
fbuf2 = framebuf.FrameBuffer(buf2, w2, h2, framebuf.GS2_HMSB)

# blit
fbuf2.fill(0)
fbuf2.pixel(0, 0, 1)
fbuf2.pixel(0, 2, 2)
fbuf2.pixel(1, 0, 1)
fbuf2.pixel(1, 2, 2)
fbuf.fill(3)
fbuf.blit(fbuf2, 3, 3, 0)
fbuf.blit(fbuf2, -1, -1, 0)
fbuf.blit(fbuf2, 16, 16, 0)
printbuf()



# Test true-ish value handling

if not False:
    print("False")

if not None:
    print("None")

if not 0:
    print("0")

if not "":
    print("Empty string")
if "foo":
    print("Non-empty string")

if not ():
    print("Empty tuple")
if ("",):
    print("Non-empty tuple")

if not []:
    print("Empty list")
if [0]:
    print("Non-empty list")

if not {}:
    print("Empty dict")
if {0:0}:
    print("Non-empty dict")



# test calling a function with keywords given by **dict

def f(a, b):
    print(a, b)

f(1, **{'b':2})
f(1, **{'b':val for val in range(1)})

try:
    f(1, **{len:2})
except TypeError:
    print('TypeError')

# test calling a method with keywords given by **dict

class A:
    def f(self, a, b):
        print(a, b)

a = A()
a.f(1, **{'b':2})
a.f(1, **{'b':val for val in range(1)})

try:
    f(1, **{len: 1})
except TypeError:
    print(True)
else:
    print(False)



"""
1
"""

def _f(): pass
FunctionType = type(_f)
LambdaType = type(lambda: None)
CodeType = None
MappingProxyType = None
SimpleNamespace = None

def _g():
    yield 1
GeneratorType = type(_g())

class _C:
    def _m(self): pass
MethodType = type(_C()._m)

BuiltinFunctionType = type(len)
BuiltinMethodType = type([].append)

del _f

# print only the first 8 chars, since we have different str rep to CPython
print(str(FunctionType)[:8])
print(str(BuiltinFunctionType)[:8])



# test handling of failed heap allocation with bytearray

import micropython


class GetSlice:
    def __getitem__(self, idx):
        return idx


sl = GetSlice()[:]

# create bytearray
micropython.heap_lock()
try:
    bytearray(4)
except MemoryError:
    print("MemoryError: bytearray create")
micropython.heap_unlock()

# create bytearray from bytes
micropython.heap_lock()
try:
    bytearray(b"0123")
except MemoryError:
    print("MemoryError: bytearray create from bytes")
micropython.heap_unlock()

# create bytearray from iterator
r = range(4)
micropython.heap_lock()
try:
    bytearray(r)
except MemoryError:
    print("MemoryError: bytearray create from iter")
micropython.heap_unlock()

# bytearray add
b = bytearray(4)
micropython.heap_lock()
try:
    b + b"01"
except MemoryError:
    print("MemoryError: bytearray.__add__")
micropython.heap_unlock()

# bytearray iadd
b = bytearray(4)
micropython.heap_lock()
try:
    b += b"01234567"
except MemoryError:
    print("MemoryError: bytearray.__iadd__")
micropython.heap_unlock()
print(b)

# bytearray append
b = bytearray(4)
micropython.heap_lock()
try:
    for i in range(100):
        b.append(1)
except MemoryError:
    print("MemoryError: bytearray.append")
micropython.heap_unlock()

# bytearray extend
b = bytearray(4)
micropython.heap_lock()
try:
    b.extend(b"01234567")
except MemoryError:
    print("MemoryError: bytearray.extend")
micropython.heap_unlock()

# bytearray get with slice
b = bytearray(4)
micropython.heap_lock()
try:
    b[sl]
except MemoryError:
    print("MemoryError: bytearray subscr get")
micropython.heap_unlock()

# extend bytearray using slice subscr
b = bytearray(4)
micropython.heap_lock()
try:
    b[sl] = b"01234567"
except MemoryError:
    print("MemoryError: bytearray subscr grow")
micropython.heap_unlock()
print(b)



import array


@micropython.asm_thumb  # test vldr, vstr
def arrayadd(r0):
    vldr(s0, [r0, 0])
    vldr(s1, [r0, 4])
    vadd(s2, s0, s1)
    vstr(s2, [r0, 8])


z = array.array("f", [2, 4, 10])
arrayadd(z)
print(z[2])



# MicroPython uasyncio module
# MIT license; Copyright (c) 2019-2020 Damien P. George

from . import core


async def wait_for(aw, timeout, sleep=core.sleep):
    aw = core._promote_to_task(aw)
    if timeout is None:
        return await aw

    def runner(waiter, aw):
        nonlocal status, result
        try:
            result = await aw
            s = True
        except BaseException as er:
            s = er
        if status is None:
            # The waiter is still waiting, set status for it and cancel it.
            status = s
            waiter.cancel()

    # Run aw in a separate runner task that manages its exceptions.
    status = None
    result = None
    runner_task = core.create_task(runner(core.cur_task, aw))

    try:
        # Wait for the timeout to elapse.
        await sleep(timeout)
    except core.CancelledError as er:
        if status is True:
            # aw completed successfully and cancelled the sleep, so return aw's result.
            return result
        elif status is None:
            # This wait_for was cancelled externally, so cancel aw and re-raise.
            status = True
            runner_task.cancel()
            raise er
        else:
            # aw raised an exception, propagate it out to the caller.
            raise status

    # The sleep finished before aw, so cancel aw and raise TimeoutError.
    status = True
    runner_task.cancel()
    await runner_task
    raise core.TimeoutError


def wait_for_ms(aw, timeout):
    return wait_for(aw, timeout, core.sleep_ms)


async def gather(*aws, return_exceptions=False):
    ts = [core._promote_to_task(aw) for aw in aws]
    for i in range(len(ts)):
        try:
            # TODO handle cancel of gather itself
            # if ts[i].coro:
            #    iter(ts[i]).waiting.push_head(cur_task)
            #    try:
            #        yield
            #    except CancelledError as er:
            #        # cancel all waiting tasks
            #        raise er
            ts[i] = await ts[i]
        except Exception as er:
            if return_exceptions:
                ts[i] = er
            else:
                raise er
    return ts



import jni

try:
    Integer = jni.cls("java/lang/Integer")
except:
    print("SKIP")
    raise SystemExit

# Create object
i = Integer(42)
print(i)
# Call object method
print(i.hashCode())
# Pass object to another method
System = jni.cls("java/lang/System")
System.out.println(i)



# Compute the Mandelbrot set, to test complex numbers


def mandelbrot(w, h):
    def in_set(c):
        z = 0
        for i in range(32):
            z = z * z + c
            if abs(z) > 100:
                return i
        return 0

    img = bytearray(w * h)

    xscale = (w - 1) / 2.4
    yscale = (h - 1) / 3.2
    for v in range(h):
        line = memoryview(img)[v * w : v * w + w]
        for u in range(w):
            c = in_set(complex(v / yscale - 2.3, u / xscale - 1.2))
            line[u] = c

    return img


bm_params = {
    (100, 100): (20, 20),
    (1000, 1000): (80, 80),
    (5000, 1000): (150, 150),
}


def bm_setup(ps):
    return lambda: mandelbrot(ps[0], ps[1]), lambda: (ps[0] * ps[1], None)



# This tests long ints for 32-bit machine

a = 0x1ffffffff
b = 0x100000000
print(a)
print(b)
print(a + b)
print(a - b)
print(b - a)
# overflows long long implementation
#print(a * b)
print(a // b)
print(a % b)
print("&", a & b)
print(a | b)
print(a ^ b)
print(a << 3)
print(a >> 1)

a += b
print(a)
a -= 123456
print(a)
a *= 257
print(a)
a //= 257
print(a)
a %= b
print(a)
a ^= b
print(a)
a |= b
print(a)
a &= b
print("&=", a)
a <<= 5
print(a)
a >>= 1
print(a)

# Test referential integrity of long ints
a = 0x1ffffffff
b = a
a += 1
print(a)
print(b)

# Bitwise ops on 64-bit

a = 0x1ffffffffffffffff
b = 0x10000000000000000
print("&", a & b)
print(a | b)
print(a ^ b)



# basic tests for gc module

try:
    import gc
except ImportError:
    print("SKIP")
    raise SystemExit

print(gc.isenabled())
gc.disable()
print(gc.isenabled())
gc.enable()
print(gc.isenabled())

gc.collect()

if hasattr(gc, 'mem_free'):
    # uPy has these extra functions
    # just test they execute and return an int
    assert type(gc.mem_free()) is int
    assert type(gc.mem_alloc()) is int

if hasattr(gc, 'threshold'):
    # uPy has this extra function
    # check execution and returns
    assert(gc.threshold(1) is None)
    assert(gc.threshold() == 0)
    assert(gc.threshold(-1) is None)
    assert(gc.threshold() == -1)

    # Setting a low threshold should trigger collection at the list alloc
    gc.threshold(1)
    [[], []]
    gc.threshold(-1)



# using False as a conditional
@micropython.viper
def f():
    x = False
    if x:
        pass
    else:
        print("not x", x)


f()

# using True as a conditional
@micropython.viper
def f():
    x = True
    if x:
        print("x", x)


f()

# using an int as a conditional
@micropython.viper
def g():
    y = 1
    if y:
        print("y", y)


g()

# using an int as a conditional that has the lower 16-bits clear
@micropython.viper
def h():
    z = 0x10000
    if z:
        print("z", z)


h()



# to test keyword-only arguments

# simplest case
def f(*, a):
    print(a)

f(a=1)

# with 2 keyword-only args
def f(*, a, b):
    print(a, b)

f(a=1, b=2)
f(b=1, a=2)

# positional followed by bare star
def f(a, *, b, c):
    print(a, b, c)

f(1, b=3, c=4)
f(1, c=3, b=4)
f(1, **{'b':'3', 'c':4})

try:
    f(1)
except TypeError:
    print("TypeError")

try:
    f(1, b=2)
except TypeError:
    print("TypeError")

try:
    f(1, c=2)
except TypeError:
    print("TypeError")

# with **kw
def f(a, *, b, **kw):
    print(a, b, kw)

f(1, b=2)
f(1, b=2, c=3)

# with named star
def f(*a, b, c):
    print(a, b, c)

f(b=1, c=2)
f(c=1, b=2)

# with positional and named star
def f(a, *b, c):
    print(a, b, c)

f(1, c=2)
f(1, 2, c=3)
f(a=1, c=3)

# lambda as kw-only arg (exposes nested behaviour in compiler)
def f(*, x=lambda:1):
    return x()
print(f())
print(f(x=f))
print(f(x=lambda:2))



# Source: https://github.com/python/pyperformance
# License: MIT

# Calculating some of the digits of π.
# This benchmark stresses big integer arithmetic.
# Adapted from code on: http://benchmarksgame.alioth.debian.org/


def compose(a, b):
    aq, ar, as_, at = a
    bq, br, bs, bt = b
    return (aq * bq, aq * br + ar * bt, as_ * bq + at * bs, as_ * br + at * bt)


def extract(z, j):
    q, r, s, t = z
    return (q * j + r) // (s * j + t)


def gen_pi_digits(n):
    z = (1, 0, 0, 1)
    k = 1
    digs = []
    for _ in range(n):
        y = extract(z, 3)
        while y != extract(z, 4):
            z = compose(z, (k, 4 * k + 2, 0, 2 * k + 1))
            k += 1
            y = extract(z, 3)
        z = compose((10, -10 * y, 0, 1), z)
        digs.append(y)
    return digs


###########################################################################
# Benchmark interface

bm_params = {
    (50, 25): (1, 35),
    (100, 100): (1, 65),
    (1000, 1000): (2, 250),
    (5000, 1000): (3, 350),
}


def bm_setup(params):
    state = None

    def run():
        nonlocal state
        nloop, ndig = params
        ndig = params[1]
        for _ in range(nloop):
            state = None  # free previous result
            state = gen_pi_digits(ndig)

    def result():
        return params[0] * params[1], "".join(str(d) for d in state)

    return run, result



# test errors from bad operations (unary, binary, etc)

# unsupported unary operators
try:
    ~bytearray()
except TypeError:
    print('TypeError')

# unsupported binary operators
try:
    bytearray() // 2
except TypeError:
    print('TypeError')

# object with buffer protocol needed on rhs
try:
    bytearray(1) + 1
except TypeError:
    print('TypeError')



# Types are hashable
print(hash(type) != 0)
print(hash(int) != 0)
print(hash(list) != 0)
class Foo: pass
print(hash(Foo) != 0)

print(int == int)
print(int != list)

d = {}
d[int] = list
d[list] = int
print(len(d))



# test deeply recursive generators

# simple "yield from" recursion
def gen():
    yield from gen()


try:
    list(gen())
except RuntimeError:
    print("RuntimeError")

# recursion via an iterator over a generator
def gen2():
    for x in gen2():
        yield x


try:
    next(gen2())
except RuntimeError:
    print("RuntimeError")



# This test is exactly like uctypes_le.py, but uses native structure layout.
# Codepaths for packed vs native structures are different. This test only works
# on little-endian machine (no matter if 32 or 64 bit).
import sys

try:
    import uctypes
except ImportError:
    print("SKIP")
    raise SystemExit

if sys.byteorder != "little":
    print("SKIP")
    raise SystemExit


desc = {
    "s0": uctypes.UINT16 | 0,
    "sub": (
        0,
        {
            "b0": uctypes.UINT8 | 0,
            "b1": uctypes.UINT8 | 1,
        },
    ),
    "arr": (uctypes.ARRAY | 0, uctypes.UINT8 | 2),
    "arr2": (uctypes.ARRAY | 0, 2, {"b": uctypes.UINT8 | 0}),
    "bitf0": uctypes.BFUINT16 | 0 | 0 << uctypes.BF_POS | 8 << uctypes.BF_LEN,
    "bitf1": uctypes.BFUINT16 | 0 | 8 << uctypes.BF_POS | 8 << uctypes.BF_LEN,
    "bf0": uctypes.BFUINT16 | 0 | 0 << uctypes.BF_POS | 4 << uctypes.BF_LEN,
    "bf1": uctypes.BFUINT16 | 0 | 4 << uctypes.BF_POS | 4 << uctypes.BF_LEN,
    "bf2": uctypes.BFUINT16 | 0 | 8 << uctypes.BF_POS | 4 << uctypes.BF_LEN,
    "bf3": uctypes.BFUINT16 | 0 | 12 << uctypes.BF_POS | 4 << uctypes.BF_LEN,
    "ptr": (uctypes.PTR | 0, uctypes.UINT8),
    "ptr2": (uctypes.PTR | 0, {"b": uctypes.UINT8 | 0}),
}

data = bytearray(b"01")

S = uctypes.struct(uctypes.addressof(data), desc, uctypes.NATIVE)

# print(S)
print(hex(S.s0))
assert hex(S.s0) == "0x3130"
# print(S.sub.b0)
print(S.sub.b0, S.sub.b1)
assert S.sub.b0, S.sub.b1 == (0x30, 0x31)

try:
    S[0]
    assert False, "Can't index struct"
except TypeError:
    print("TypeError")

print("arr:", S.arr[0], S.arr[1])
assert (S.arr[0], S.arr[1]) == (0x30, 0x31)

print("arr of struct:", S.arr2[0].b, S.arr2[1].b)
assert (S.arr2[0].b, S.arr2[1].b) == (0x30, 0x31)


try:
    S.arr[2]
    assert False, "Out of bounds index"
except IndexError:
    print("IndexError")

print("bf:", S.bitf0, S.bitf1)
assert (S.bitf0, S.bitf1) == (0x30, 0x31)

print("bf 4bit:", S.bf3, S.bf2, S.bf1, S.bf0)
assert (S.bf3, S.bf2, S.bf1, S.bf0) == (3, 1, 3, 0)

# Write access

S.sub.b0 = ord("2")
print(data)
assert bytes(data) == b"21"

S.bf3 = 5
print(data)
assert bytes(data) == b"2Q"

desc2 = {
    "bf8": uctypes.BFUINT8 | 0 | 0 << uctypes.BF_POS | 4 << uctypes.BF_LEN,
    "bf32": uctypes.BFUINT32 | 0 | 20 << uctypes.BF_POS | 4 << uctypes.BF_LEN,
}

data2 = bytearray(b"0123")

S2 = uctypes.struct(uctypes.addressof(data2), desc2, uctypes.NATIVE)

# bitfield using uint8 as base type
S2.bf8 = 5
print(data2)
assert bytes(data2) == b"5123"

# bitfield using uint32 as base type
S2.bf32 = 5
print(data2)
assert bytes(data2) == b"51R3"



try:
    from array import array
except ImportError:
    print("SKIP")
    raise SystemExit

def test_array_overflow(typecode, val):
    try:
        print(array(typecode, [val]))
    except OverflowError:
        print('OverflowError')

def test_bytearray_overflow(val):
    try:
        print(bytearray([val]))
    except (OverflowError, ValueError):
        # CircuitPython always does OverflowError
        print('(OverflowError, ValueError)')


# small int -1
test_array_overflow('Q', -1)
test_array_overflow('L', -1)
test_array_overflow('I', -1)
test_array_overflow('H', -1)
test_array_overflow('B', -1)

# 0 ok
test_array_overflow('Q', 0)
test_array_overflow('L', 0)
test_array_overflow('I', 0)
test_array_overflow('H', 0)
test_array_overflow('B', 0)

# 1 ok
test_array_overflow('Q', 1)
test_array_overflow('L', 1)
test_array_overflow('I', 1)
test_array_overflow('H', 1)
test_array_overflow('B', 1)

# truth value conversions
test_array_overflow('b', True)
test_array_overflow('b', False)

# similar tests for bytearrays
test_bytearray_overflow(0)
test_bytearray_overflow(1)
test_bytearray_overflow(-1)
test_bytearray_overflow(256)
test_bytearray_overflow(True)
test_bytearray_overflow(False)



# Regression test for #290 - throwing exception in another module led to
# its namespace stick and namespace of current module not coming back.
import import1b


def func1():
    print("func1")


def func2():
    try:
        import1b.throw()
    except ValueError:
        pass
    func1()


func2()



# This tests small int range for 32-bit machine

# Small ints are variable-length encoded in MicroPython, so first
# test that encoding works as expected.

print(0)
print(1)
print(-1)
# Value is split in 7-bit "subwords", and taking into account that all
# ints in Python are signed, there're 6 bits of magnitude. So, around 2^6
# there's "turning point"
print(63)
print(64)
print(65)
print(-63)
print(-64)
print(-65)
# Maximum values of small ints on 32-bit platform
print(1073741823)
# Per python semantics, lexical integer is without a sign (i.e. positive)
# and '-' is unary minus operation applied to it. That's why -1073741824
# (min two-complement's negative value) is not allowed.
print(-1073741823)

# Operations tests

# compile-time constexprs
print(1 + 3)
print(3 - 2)
print(2 * 3)
print(1 & 3)
print(1 | 2)
print(1 ^ 3)
print(+3)
print(-3)
print(~3)

a = 0x3fffff
print(a)
a *= 0x10
print(a)
a *= 0x10
print(a)
a += 0xff
print(a)
# This would overflow
#a += 1

a = -0x3fffff
print(a)
a *= 0x10
print(a)
a *= 0x10
print(a)
a -= 0xff
print(a)
# This still doesn't overflow
a -= 1
print(a)
# This would overflow
#a -= 1

# negative shifts are not allowed
try:
    a << -1
except ValueError:
    print("ValueError")
try:
    a >> -1
except ValueError:
    print("ValueError")

# Shifts to big amounts are undefined behavior in C and is CPU-specific

# These are compile-time constexprs
print(1 >> 32)
print(1 >> 64)
print(1 >> 128)

# These are runtime calcs
a = 1
print(a >> 32)
print(a >> 64)
print(a >> 128)



class A:
    def __init__(self, x):
        print('A init', x)
        self.x = x

    def f(self):
        print(self.x, self.y)

class B(A):
    def __init__(self, x, y):
        A.__init__(self, x)
        print('B init', x, y)
        self.y = y

    def g(self):
        print(self.x, self.y)

A(1)
b = B(1, 2)
b.f()
b.g()



# Test cancelling a task

try:
    import uasyncio as asyncio
except ImportError:
    try:
        import asyncio
    except ImportError:
        print("SKIP")
        raise SystemExit


async def task(s, allow_cancel):
    try:
        print("task start")
        await asyncio.sleep(s)
        print("task done")
    except asyncio.CancelledError as er:
        print("task cancel")
        if allow_cancel:
            raise er


async def task2(allow_cancel):
    print("task 2")
    try:
        await asyncio.create_task(task(0.05, allow_cancel))
    except asyncio.CancelledError as er:
        print("task 2 cancel")
        raise er
    print("task 2 done")


async def main():
    # Cancel task immediately
    t = asyncio.create_task(task(2, True))
    print(t.cancel())

    # Cancel task after it has started
    t = asyncio.create_task(task(2, True))
    await asyncio.sleep(0.01)
    print(t.cancel())
    print("main sleep")
    await asyncio.sleep(0.01)

    # Cancel task multiple times after it has started
    t = asyncio.create_task(task(2, True))
    await asyncio.sleep(0.01)
    for _ in range(4):
        print(t.cancel())
    print("main sleep")
    await asyncio.sleep(0.01)

    # Await on a cancelled task
    print("main wait")
    try:
        await t
    except asyncio.CancelledError:
        print("main got CancelledError")

    # Cancel task after it has finished
    t = asyncio.create_task(task(0.01, False))
    await asyncio.sleep(0.05)
    print(t.cancel())

    # Nested: task2 waits on task, task2 is cancelled (should cancel task then task2)
    print("----")
    t = asyncio.create_task(task2(True))
    await asyncio.sleep(0.01)
    print("main cancel")
    t.cancel()
    print("main sleep")
    await asyncio.sleep(0.1)

    # Nested: task2 waits on task, task2 is cancelled but task doesn't allow it (task2 should continue)
    print("----")
    t = asyncio.create_task(task2(False))
    await asyncio.sleep(0.01)
    print("main cancel")
    t.cancel()
    print("main sleep")
    await asyncio.sleep(0.1)


asyncio.run(main())



class Cud():

    def __init__(self):
        print("__init__ called")

    def __repr__(self):
        print("__repr__ called")
        return ""

    def __lt__(self, other):
        print("__lt__ called")

    def __le__(self, other):
        print("__le__ called")

    def __eq__(self, other):
        print("__eq__ called")

    def __ne__(self, other):
        print("__ne__ called")

    def __ge__(self, other):
        print("__ge__ called")

    def __gt__(self, other):
        print("__gt__ called")

    def __abs__(self):
        print("__abs__ called")

    def __add__(self, other):
        print("__add__ called")

    def __and__(self, other):
        print("__and__ called")

    def __floordiv__(self, other):
        print("__floordiv__ called")

    def __index__(self, other):
        print("__index__ called")

    def __inv__(self):
        print("__inv__ called")

    def __invert__(self):
        print("__invert__ called")

    def __lshift__(self, val):
        print("__lshift__ called")

    def __mod__(self, val):
        print("__mod__ called")

    def __mul__(self, other):
        print("__mul__ called")

    def __matmul__(self, other):
        print("__matmul__ called")

    def __neg__(self):
        print("__neg__ called")

    def __or__(self, other):
        print("__or__ called")

    def __pos__(self):
        print("__pos__ called")

    def __pow__(self, val):
        print("__pow__ called")

    def __rshift__(self, val):
        print("__rshift__ called")

    def __sub__(self, other):
        print("__sub__ called")

    def __truediv__(self, other):
        print("__truediv__ called")

    def __div__(self, other):
        print("__div__ called")

    def __xor__(self, other):
        print("__xor__ called")

    def __iadd__(self, other):
        print("__iadd__ called")
        return self

    def __isub__(self, other):
        print("__isub__ called")
        return self

    def __int__(self):
        return 42

cud1 = Cud()
cud2 = Cud()

str(cud1)
cud1 == cud1
cud1 == cud2
cud1 != cud1
cud1 != cud2
cud1 < cud2
cud1 <= cud2
cud1 == cud2
cud1 >= cud2
cud1 > cud2
cud1 + cud2
cud1 - cud2
print(int(cud1))

class BadInt:
    def __int__(self):
        print("__int__ called")
        return None

try:
    int(BadInt())
except TypeError:
    print("TypeError")

# more in special_methods2.py



# test exception matching against a tuple

try:
    fail
except (Exception,):
    print('except 1')

try:
    fail
except (Exception, Exception):
    print('except 2')

try:
    fail
except (TypeError, NameError):
    print('except 3')

try:
    fail
except (TypeError, ValueError, Exception):
    print('except 4')



import skip_if
skip_if.no_reverse_ops()

class A:

    def __init__(self, v):
        self.v = v

    def __add__(self, o):
        if isinstance(o, A):
            return A(self.v + o.v)
        return A(self.v + o)

    def __radd__(self, o):
        return A(self.v + o)

    def __repr__(self):
        return "A({})".format(self.v)

print(A(3) + 1)
print(2 + A(5))



# test await expression

# uPy allows normal generators to be awaitables.
# CircuitPython does not.
# In CircuitPython you need to have an __await__ method on an awaitable like in CPython;
#  and like in CPython, generators do not have __await__.

class Awaitable:
    def __init__(self, value):
        self.value = value

    def __await__(self):
        print('wait value:', self.value)
        msg = yield 'message from wait({})'.format(self.value)
        print('wait got back:', msg)
        return 10

async def f():
    x = await Awaitable(1)**2
    print('x =', x)

coro = f()
print('return from send:', coro.send(None))
try:
    coro.send('message from main')
except StopIteration:
    print('got StopIteration')



# Test for VfsLfs using a RAM device, when the first superblock does not exist

try:
    import uos

    uos.VfsLfs2
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class RAMBlockDevice:
    def __init__(self, block_size, data):
        self.block_size = block_size
        self.data = data

    def readblocks(self, block, buf, off):
        addr = block * self.block_size + off
        for i in range(len(buf)):
            buf[i] = self.data[addr + i]

    def ioctl(self, op, arg):
        if op == 4:  # block count
            return len(self.data) // self.block_size
        if op == 5:  # block size
            return self.block_size
        if op == 6:  # erase block
            return 0


# This is a valid littlefs2 filesystem with a block size of 64 bytes.
# The first block (where the first superblock is stored) is fully erased.
lfs2_data = b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x02\x00\x00\x00\xf0\x0f\xff\xf7littlefs/\xe0\x00\x10\x00\x00\x02\x00@\x00\x00\x00\x04\x00\x00\x00\xff\x00\x00\x00\xff\xff\xff\x7f\xfe\x03\x00\x00p\x1f\xfc\x08\x1b\xb4\x14\xa7\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01\x00\x00\x00\xff\xef\xff\xf7test.txt \x00\x00\x08p\x1f\xfc\x08\x83\xf1u\xba\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"

# Create the block device from the static data (it will be read-only).
bdev = RAMBlockDevice(64, lfs2_data)

# Create the VFS explicitly, no auto-detection is needed for this.
vfs = uos.VfsLfs2(bdev)
print(list(vfs.ilistdir()))

# Mount the block device directly; this relies on auto-detection.
uos.mount(bdev, "/userfs")
print(uos.listdir("/userfs"))

# Clean up.
uos.umount("/userfs")



#
# This is minimal MicroPython variant of run-tests.py script, which uses
# .exp files as generated by run-tests.py --write-exp. It is useful to run
# testsuite on systems which have neither CPython3 nor unix shell.
# This script is intended to be run by the same interpreter executable
# which is to be tested, so should use minimal language functionality.
#
import sys
import uos as os


tests = ["basics", "micropython", "float", "import", "io", " misc", "unicode", "extmod", "unix"]

if sys.platform == "win32":
    MICROPYTHON = "micropython.exe"
else:
    MICROPYTHON = "micropython"


def should_skip(test):
    if test.startswith("native"):
        return True
    if test.startswith("viper"):
        return True


test_count = 0
passed_count = 0
skip_count = 0

for suite in tests:
    # print("Running in: %s" % suite)
    if sys.platform == "win32":
        # dir /b prints only contained filenames, one on a line
        # http://www.microsoft.com/resources/documentation/windows/xp/all/proddocs/en-us/dir.mspx
        r = os.system("dir /b %s/*.py >tests.lst" % suite)
    else:
        r = os.system("ls %s/*.py | xargs -n1 basename >tests.lst" % suite)
    assert r == 0

    with open("tests.lst") as f:
        testcases = f.readlines()
        testcases = [l[:-1] for l in testcases]
    assert testcases, "No tests found in dir '%s', which is implausible" % suite
    # print(testcases)
    for t in testcases:
        if t == "native_check.py":
            continue

        qtest = "%s/%s" % (suite, t)

        if should_skip(t):
            print("skip " + qtest)
            skip_count += 1
            continue

        exp = None
        try:
            f = open(qtest + ".exp")
            exp = f.read()
            f.close()
        except OSError:
            pass

        if exp is not None:
            # print("run " + qtest)
            r = os.system(MICROPYTHON + " %s >.tst.out" % qtest)
            if r == 0:
                f = open(".tst.out")
                out = f.read()
                f.close()
            else:
                out = "CRASH"

            if out == "SKIP\n":
                print("skip " + qtest)
                skip_count += 1
            else:
                if out == exp:
                    print("pass " + qtest)
                    passed_count += 1
                else:
                    print("FAIL " + qtest)

                test_count += 1
        else:
            skip_count += 1

print("%s tests performed" % test_count)
print("%s tests passed" % passed_count)
if test_count != passed_count:
    print("%s tests failed" % (test_count - passed_count))
if skip_count:
    print("%s tests skipped" % skip_count)



# Test that errno's propagate correctly through btree module.

try:
    import btree, uio, uerrno

    uio.IOBase
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class Device(uio.IOBase):
    def __init__(self, read_ret=0, ioctl_ret=0):
        self.read_ret = read_ret
        self.ioctl_ret = ioctl_ret

    def readinto(self, buf):
        print("read", len(buf))
        return self.read_ret

    def ioctl(self, cmd, arg):
        print("ioctl", cmd)
        return self.ioctl_ret


# Invalid pagesize; errno comes from btree library
try:
    db = btree.open(Device(), pagesize=511)
except OSError as er:
    print("OSError", er.errno == uerrno.EINVAL)

# Valid pagesize, device returns error on read; errno comes from Device.readinto
try:
    db = btree.open(Device(-1000), pagesize=512)
except OSError as er:
    print(repr(er))

# Valid pagesize, device returns error on seek; errno comes from Device.ioctl
try:
    db = btree.open(Device(0, -1001), pagesize=512)
except OSError as er:
    print(repr(er))



# test formatting floats with large precision, that it doesn't overflow the buffer


def test(num, num_str):
    if num == float("inf") or num == 0.0 and num_str != "0.0":
        # skip numbers that overflow or underflow the FP precision
        return
    for kind in ("e", "f", "g"):
        # check precision either side of the size of the buffer (32 bytes)
        for prec in range(23, 36, 2):
            fmt = "%." + "%d" % prec + kind
            s = fmt % num
            check = abs(float(s) - num)
            if num > 1:
                check /= num
            if check > 1e-6:
                print("FAIL", num_str, fmt, s, len(s), check)


# check most powers of 10, making sure to include exponents with 3 digits
for e in range(-101, 102):
    num = pow(10, e)
    test(num, "1e%d" % e)



# Test fairness of Event.set()
# That tasks which continuously wait on events don't take over the scheduler

try:
    import uasyncio as asyncio
except ImportError:
    try:
        import asyncio
    except ImportError:
        print("SKIP")
        raise SystemExit


async def task1(id):
    for i in range(4):
        print("sleep", id)
        await asyncio.sleep(0)


async def task2(id, ev):
    for i in range(4):
        ev.set()
        ev.clear()
        print("wait", id)
        await ev.wait()


async def main():
    ev = asyncio.Event()
    tasks = [
        asyncio.create_task(task1(0)),
        asyncio.create_task(task2(2, ev)),
        asyncio.create_task(task1(1)),
        asyncio.create_task(task2(3, ev)),
    ]
    await tasks[1]
    ev.set()


asyncio.run(main())



"""
categories: Core,Classes
description: When inheriting from multiple classes super() only calls one class
cause: See :ref:`cpydiff_core_class_mro`
workaround: See :ref:`cpydiff_core_class_mro`
"""


class A:
    def __init__(self):
        print("A.__init__")


class B(A):
    def __init__(self):
        print("B.__init__")
        super().__init__()


class C(A):
    def __init__(self):
        print("C.__init__")
        super().__init__()


class D(B, C):
    def __init__(self):
        print("D.__init__")
        super().__init__()


D()



try:
    import urandom as random
except ImportError:
    try:
        import random
    except ImportError:
        print("SKIP")
        raise SystemExit

try:
    random.randint
except AttributeError:
    print("SKIP")
    raise SystemExit

print("random")
for i in range(50):
    assert 0 <= random.random() < 1

print("uniform")
for i in range(50):
    assert 0 <= random.uniform(0, 4) <= 4
    assert 2 <= random.uniform(2, 6) <= 6
    assert -2 <= random.uniform(-2, 2) <= 2



# test ffi float support
try:
    import ffi
except ImportError:
    print("SKIP")
    raise SystemExit


def ffi_open(names):
    err = None
    for n in names:
        try:
            mod = ffi.open(n)
            return mod
        except OSError as e:
            err = e
    raise err


libc = ffi_open(("libc.so", "libc.so.0", "libc.so.6", "libc.dylib"))

try:
    strtof = libc.func("f", "strtof", "sp")
except OSError:
    # Some libc's (e.g. Android's Bionic) define strtof as macro/inline func
    # in terms of strtod().
    print("SKIP")
    raise SystemExit

print("%.6f" % strtof("1.23", None))

strtod = libc.func("d", "strtod", "sp")
print("%.6f" % strtod("1.23", None))

# test passing double and float args
libm = ffi_open(("libm.so", "libm.so.6", "libc.so.0", "libc.so.6", "libc.dylib"))
tgamma = libm.func("d", "tgamma", "d")
for fun_name in ("tgamma",):
    fun = globals()[fun_name]
    for val in (0.5, 1, 1.0, 1.5, 4, 4.0):
        print(fun_name, "%.5f" % fun(val))

# test passing 2x float/double args
powf = libm.func("f", "powf", "ff")
pow = libm.func("d", "pow", "dd")
for fun_name in ("powf", "pow"):
    fun = globals()[fun_name]
    for args in ((0, 1), (1, 0), (2, 0.5), (3, 4)):
        print(fun_name, "%.5f" % fun(*args))



"""
categories: Modules,array
description: Comparison between different typecodes not supported
cause: Code size
workaround: Compare individual elements
"""
import array

array.array("b", [1, 2]) == array.array("i", [1, 2])



import micropython

# Tests both code paths for built-in exception raising.
# mp_obj_new_exception_msg_varg (exception requires decompression at raise-time to format)
# mp_obj_new_exception_msg (decompression can be deferred)

# NameError uses mp_obj_new_exception_msg_varg for NameError("name '%q' isn't defined")
# set.pop uses mp_obj_new_exception_msg for KeyError("pop from an empty set")

# Tests that deferred decompression works both via print(e) and accessing the message directly via e.args.

a = set()

# First test the regular case (can use heap for allocating the decompression buffer).
try:
    name()
except NameError as e:
    print(type(e).__name__, e)

try:
    a.pop()
except KeyError as e:
    print(type(e).__name__, e)

try:
    name()
except NameError as e:
    print(e.args[0])

try:
    a.pop()
except KeyError as e:
    print(e.args[0])

# Then test that it still works when the heap is locked (i.e. in ISR context).
micropython.heap_lock()

try:
    name()
except NameError as e:
    print(type(e).__name__)

try:
    a.pop()
except KeyError as e:
    print(type(e).__name__)

micropython.heap_unlock()



# SPDX-FileCopyrightText: 2014 MicroPython & CircuitPython contributors (https://github.com/adafruit/circuitpython/graphs/contributors)
#
# SPDX-License-Identifier: MIT

# Validate that all entries in the .pot are in every .po. Only the .pot is updated so we can detect
# if a translation was added to the source but isn't in a .po. This ensures translators can grab
# complete files to work on.

import git
import sys
import polib

po_filenames = sys.argv[1:]
repo = git.Repo()

NO_TRANSLATION_WHITELIST = ["ffi_prep_closure_loc"]

bad_commits = {}
for po_filename in po_filenames:
    print("Checking", po_filename)

    commits = repo.iter_commits(paths=po_filename, reverse=True, topo_order=True)
    first_translations = None
    fixed_ids = set()
    for commit in commits:
        try:
            blob = commit.tree / po_filename
        except KeyError:
            continue
        try:
            current_file = polib.pofile(blob.data_stream.read().decode("utf-8"))
        except OSError:
            print("skipping invalid po in", commit)
            continue
        if not first_translations:
            first_translations = current_file
            continue
        print(commit.authored_date, commit)
        first_translations.metadata = current_file.metadata
        first_translations.header = current_file.header
        for entry in first_translations:
            newer_entry = current_file.find(entry.msgid)

            if newer_entry and entry.msgid == "":
                entry.merge(newer_entry)
                print(entry)
            elif newer_entry and entry.msgstr != newer_entry.msgstr:
                if newer_entry.msgstr != "" and (
                    newer_entry.msgstr != entry.msgid or entry.msgid in NO_TRANSLATION_WHITELIST
                ):
                    entry.merge(newer_entry)
                    entry.msgstr = newer_entry.msgstr
                    if "fuzzy" not in newer_entry.flags and "fuzzy" in entry.flags:
                        entry.flags.remove("fuzzy")
                elif entry.msgid not in fixed_ids:
                    if commit not in bad_commits:
                        bad_commits[commit] = set()
                    bad_commits[commit].add(po_filename)
                    fixed_ids.add(entry.msgid)
                    # print(entry.msgid, "\"" + entry.msgstr + "\"", "\"" + newer_entry.msgstr + "\"",)
            elif newer_entry and newer_entry.flags != entry.flags:
                entry.flags = newer_entry.flags
            elif newer_entry and newer_entry.obsolete != entry.obsolete:
                entry.obsolete = newer_entry.obsolete
            elif not newer_entry:
                entry.obsolete = True

        # Add new entries to the modified pofile.
        for entry in current_file:
            old_entry = first_translations.find(entry.msgid, include_obsolete_entries=True)
            if old_entry:
                continue
            first_translations.append(entry)

    # Remove obsolete translations. We can always use this script to revive them from the git history.
    to_remove = []
    for entry in first_translations:
        if entry.obsolete:
            to_remove.append(entry)

    for remove in to_remove:
        first_translations.remove(remove)

    first_translations.save(po_filename)

print()
for commit in bad_commits:
    files = bad_commits[commit]
    print(commit)
    for file in files:
        print("\t", file)



# test micropython.heap_locked()

import micropython

if not hasattr(micropython, "heap_locked"):
    print("SKIP")
    raise SystemExit

micropython.heap_lock()
print(micropython.heap_locked())
micropython.heap_unlock()
print(micropython.heap_locked())



import time
import adafruit_ble
from adafruit_ble_eddystone import uid, url
import alarm

np = neopixel.NeoPixel(board.NEOPIXEL, 1)

radio = adafruit_ble.BLERadio()
# Reuse the BLE address as our Eddystone instance id.
eddystone_uid = uid.EddystoneUID(radio.address_bytes)
eddystone_url = url.EddystoneURL("https://adafru.it/discord")

while True:
    np[0] = (50, 0, 0)
    # Alternate between advertising our ID and our URL.
    radio.start_advertising(eddystone_uid)
    time.sleep(0.5)
    radio.stop_advertising()

    radio.start_advertising(eddystone_url)
    time.sleep(0.5)
    radio.stop_advertising()

    ## USB enumeration may take 4-5s per restart
    time_alarm = alarm.time.TimeAlarm(monotonic_time=time.monotonic() + 10)

    np[0] = (0, 0, 0)

    alarm.exit_and_deep_sleep_until_alarms(time_alarm)



# Test basic, stand-alone TCP socket functionality

try:
    import usocket as socket, uerrno as errno
except ImportError:
    try:
        import socket, errno
    except ImportError:
        print("SKIP")
        raise SystemExit

# recv() on a fresh socket should raise ENOTCONN
s = socket.socket()
try:
    s.recv(1)
except OSError as er:
    print("ENOTCONN:", er.errno == errno.ENOTCONN)



"""
categories: Modules,random
description: ``getrandbits`` method can only return a maximum of 32 bits at a time.
cause: PRNG's internal state is only 32bits so it can only return a maximum of 32 bits of data at a time.
workaround: If you need a number that has more than 32 bits then utilize the random module from micropython-lib.
"""

import random


x = random.getrandbits(64)
print("{}".format(x))



# Creating BytesIO from immutable object should not immediately
# copy its content.
try:
    import uio
    import micropython

    micropython.mem_total
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


data = b"1234" * 256

before = micropython.mem_total()

buf = uio.BytesIO(data)

after = micropython.mem_total()

print(after - before < len(data))



# test builtin callable

# primitives should not be callable
print(callable(None))
print(callable(1))
print(callable([]))
print(callable("dfsd"))

# modules should not be callabe
import sys
print(callable(sys))

# builtins should be callable
print(callable(callable))

# lambdas should be callable
print(callable(lambda:None))

# user defined functions should be callable
def f():
    pass
print(callable(f))

# types should be callable, but not instances
class A:
    pass
print(callable(A))
print(callable(A()))

# instances with __call__ method should be callable
class B:
    def __call__(self):
        pass
print(callable(B()))

# this checks internal use of callable when extracting members from an instance
class C:
    def f(self):
        return "A.f"
class D:
    g = C() # g is a value and is not callable
print(callable(D().g))
print(D().g.f())



# del name

x = 1
print(x)
del x
try:
    print(x)
except NameError:
    print("NameError")
try:
    del x
except: # NameError:
    # FIXME uPy returns KeyError for this
    print("NameError")

class C:
    def f():
        pass



try:
    import uasyncio as asyncio
except ImportError:
    try:
        import asyncio
    except ImportError:
        print("SKIP")
        raise SystemExit


async def forever():
    print("forever start")
    await asyncio.sleep(10)


async def main():
    print("main start")
    asyncio.create_task(forever())
    await asyncio.sleep(0.001)
    print("main done")
    return 42


print(asyncio.run(main()))



"""
categories: Modules,os
description: ``environ`` attribute is not implemented
cause: Unknown
workaround: Use ``getenv``, ``putenv`` and ``unsetenv``
"""
import os

try:
    print(os.environ.get("NEW_VARIABLE"))
    os.environ["NEW_VARIABLE"] = "VALUE"
    print(os.environ["NEW_VARIABLE"])
except AttributeError:
    print("should not get here")
    print(os.getenv("NEW_VARIABLE"))
    os.putenv("NEW_VARIABLE", "VALUE")
    print(os.getenv("NEW_VARIABLE"))



try:
    try:
        from ucollections import deque
    except ImportError:
        from collections import deque
except ImportError:
    print("SKIP")
    raise SystemExit


d = deque((), 2)
print(len(d))
print(bool(d))

try:
    d.popleft()
except IndexError:
    print("IndexError")

print(d.append(1))
print(len(d))
print(bool(d))
print(d.popleft())
print(len(d))

d.append(2)
print(d.popleft())

d.append(3)
d.append(4)
print(len(d))
print(d.popleft(), d.popleft())
try:
    d.popleft()
except IndexError:
    print("IndexError")

d.append(5)
d.append(6)
d.append(7)
print(len(d))
print(d.popleft(), d.popleft())
print(len(d))
try:
    d.popleft()
except IndexError:
    print("IndexError")

# Case where get index wraps around when appending to full deque
d = deque((), 2)
d.append(1)
d.append(2)
d.append(3)
d.append(4)
d.append(5)
print(d.popleft(), d.popleft())

# Negative maxlen is not allowed
try:
    deque((), -1)
except ValueError:
    print("ValueError")

# Unsupported unary op
try:
    ~d
except TypeError:
    print("TypeError")



# tests transition from small to large int representation by addition

# 31-bit overflow
i = 0x3fffffff
print(i + i)
print(-i + -i)

# 47-bit overflow
i = 0x3fffffffffff
print(i + i)
print(-i + -i)

# 63-bit overflow
i = 0x3fffffffffffffff
print(i + i)
print(-i + -i)



# make sure syntax error works correctly for bad const definition

from micropython import const


def test_syntax(code):
    try:
        exec(code)
    except SyntaxError:
        print("SyntaxError")


# argument not a constant
test_syntax("a = const(x)")

# redefined constant
test_syntax("A = const(1); A = const(2)")

# these operations are not supported within const
test_syntax("A = const(1 @ 2)")
test_syntax("A = const(1 / 2)")
test_syntax("A = const(1 ** -2)")
test_syntax("A = const(1 << -2)")
test_syntax("A = const(1 >> -2)")
test_syntax("A = const(1 % 0)")
test_syntax("A = const(1 // 0)")



def gen():
    i = 0
    while 1:
        yield i
        i += 1

g = gen()

try:
    g.pend_throw
except AttributeError:
    print("SKIP")
    raise SystemExit


# Verify that an injected exception will be raised from next().
print(next(g))
print(next(g))
g.pend_throw(ValueError())

v = None
try:
    v = next(g)
except Exception as e:
    print("raised", repr(e))

print("ret was:", v)


# Verify that pend_throw works on an unstarted coroutine.
g = gen()
g.pend_throw(OSError())
try:
    next(g)
except Exception as e:
    print("raised", repr(e))


# Verify that you can't resume the coroutine from within the running coroutine.
def gen_next():
    next(g)
    yield 1

g = gen_next()

try:
    next(g)
except Exception as e:
    print("raised", repr(e))


# Verify that you can't pend_throw from within the running coroutine.
def gen_pend_throw():
    g.pend_throw(ValueError())
    yield 1

g = gen_pend_throw()

try:
    next(g)
except Exception as e:
    print("raised", repr(e))


# Verify that the pend_throw exception can be ignored.
class CancelledError(Exception):
    pass

def gen_cancelled():
    for i in range(5):
        try:
            yield i
        except CancelledError:
            print('ignore CancelledError')

g = gen_cancelled()
print(next(g))
g.pend_throw(CancelledError())
print(next(g))
# ...but not if the generator hasn't started.
g = gen_cancelled()
g.pend_throw(CancelledError())
try:
    next(g)
except Exception as e:
    print("raised", repr(e))


# Verify that calling pend_throw returns the previous exception.
g = gen()
next(g)
print(repr(g.pend_throw(CancelledError())))
print(repr(g.pend_throw(OSError)))


# Verify that you can pend_throw(None) to cancel a previous pend_throw.
g = gen()
next(g)
g.pend_throw(CancelledError())
print(repr(g.pend_throw(None)))
print(next(g))



# testing default args to a function

def fun1(val=5):
    print(val)

fun1()
fun1(10)

def fun2(p1, p2=100, p3="foo"):
    print(p1, p2, p3)

fun2(1)
fun2(1, None)
fun2(0, "bar", 200)
try:
    fun2()
except TypeError:
    print("TypeError")
try:
    fun2(1, 2, 3, 4)
except TypeError:
    print("TypeError")

# lambda as default arg (exposes nested behaviour in compiler)
def f(x=lambda:1):
    return x()
print(f())
print(f(f))
print(f(lambda:2))



# test behaviour of module objects

# this module should always exist
import __main__

# print module
print(repr(__main__).startswith("<module '__main__'"))

# store new attribute
__main__.x = 1

# delete attribute
del __main__.x



# SPDX-FileCopyrightText: 2014 MicroPython & CircuitPython contributors (https://github.com/adafruit/circuitpython/graphs/contributors)
#
# SPDX-License-Identifier: MIT

# This script renders a graph of the MicroPython heap at the given point it was dumped.
# It takes three files, the binary dump of ram, the binary for CircuitPython and the linker map file.

# To dump ram do this in GDB: dump binary memory ram.bin &_srelocate &_estack

import binascii
import struct
import sys
import pygraphviz as pgv
import io
import html
import os.path
import string

import click

from analyze_mpy import Prelude

BITS_PER_BYTE = 8
BLOCKS_PER_ATB = 4
BLOCKS_PER_FTB = 8
BYTES_PER_BLOCK = 16

AT_FREE = 0
AT_HEAD = 1
AT_TAIL = 2
AT_MARK = 3

MICROPY_QSTR_BYTES_IN_HASH = 1
MICROPY_QSTR_BYTES_IN_LEN = 1

MP_OBJ_NULL = 0
MP_OBJ_SENTINEL = 4

READLINE_HIST_SIZE = 8

SKIP_SYMBOLS = [
    ".debug_ranges",
    ".debug_frame",
    ".debug_loc",
    ".comment",
    ".debug_str",
    ".debug_line",
    ".debug_abbrev",
    ".debug_info",
    "COMMON",
]


@click.command()
@click.argument("ram_filename")
@click.argument("bin_filename")
@click.argument("map_filename")
@click.option(
    "--print_block_contents", default=False, help="Prints the contents of each allocated block"
)
@click.option(
    "--print_unknown_types",
    default=False,
    help="Prints the micropython base type if we don't understand it.",
)
@click.option(
    "--print_block_state", default=False, help="Prints the heap block states (allocated or free)"
)
@click.option(
    "--print_conflicting_symbols", default=False, help="Prints conflicting symbols from the map"
)
@click.option(
    "--print-heap-structure/--no-print-heap-structure", default=False, help="Print heap structure"
)
@click.option("--output_directory", default="heapvis", help="Destination for rendered output")
@click.option(
    "--draw-heap-layout/--no-draw-heap-layout", default=True, help="Draw the heap layout"
)
@click.option(
    "--draw-heap-ownership/--no-draw-heap-ownership",
    default=False,
    help="Draw the ownership graph of blocks on the heap",
)
@click.option("--analyze-snapshots", default="last", type=click.Choice(["all", "last"]))
def do_all_the_things(
    ram_filename,
    bin_filename,
    map_filename,
    print_block_contents,
    print_unknown_types,
    print_block_state,
    print_conflicting_symbols,
    print_heap_structure,
    output_directory,
    draw_heap_layout,
    draw_heap_ownership,
    analyze_snapshots,
):
    with open(ram_filename, "rb") as f:
        ram_dump = f.read()

    with open(bin_filename, "rb") as f:
        rom = f.read()

    symbols = {}  # name -> address, size
    symbol_lookup = {}  # address -> name
    manual_symbol_map = {}  # autoname -> name

    def add_symbol(name, address=None, size=None):
        if "lto_priv" in name:
            name = name.split(".")[0]
        if address:
            address = int(address, 0)
        if size:
            size = int(size, 0)
        if name in symbols:
            if address and symbols[name][0] and symbols[name][0] != address:
                if print_conflicting_symbols:
                    print(
                        "Conflicting symbol: {} at addresses 0x{:08x} and 0x{:08x}".format(
                            name, address, symbols[name][0]
                        )
                    )
                return
            if not address:
                address = symbols[name][0]
            if not size:
                size = symbols[name][1]
        symbols[name] = (address, size)
        if address:
            if not size:
                size = 4
            for offset in range(0, size, 4):
                symbol_lookup[address + offset] = "{}+{}".format(name, offset)

    with open(map_filename, "r") as f:
        common_symbols = False
        name = None
        for line in f:
            line = line.strip()
            parts = line.split()
            if line.startswith("Common symbol"):
                common_symbols = True
            if line == "Discarded input sections":
                common_symbols = False
            if common_symbols:
                if len(parts) == 1:
                    name = parts[0]
                elif len(parts) == 2 and name:
                    add_symbol(name, size=parts[0])
                    name = None
                elif len(parts) == 3:
                    add_symbol(parts[0], size=parts[1])
                    name = None
            else:
                if (
                    len(parts) == 1
                    and parts[0].startswith((".text", ".rodata", ".bss"))
                    and parts[0].count(".") > 1
                    and not parts[0].isnumeric()
                    and ".str" not in parts[0]
                ):
                    name = parts[0].split(".")[2]
                if (
                    len(parts) == 3
                    and parts[0].startswith("0x")
                    and parts[1].startswith("0x")
                    and name
                ):
                    add_symbol(name, parts[0], parts[1])
                    name = None
                if len(parts) == 2 and parts[0].startswith("0x") and not parts[1].startswith("0x"):
                    add_symbol(parts[1], parts[0])
                if (
                    len(parts) == 4
                    and parts[0] not in SKIP_SYMBOLS
                    and parts[1].startswith("0x")
                    and parts[2].startswith("0x")
                ):
                    name, address, size, source = parts
                    if name.startswith((".text", ".rodata", ".bss")) and name.count(".") > 1:
                        name = name.split(".")[-1]
                        add_symbol(name, address, size)
                    name = None
                # Linker symbols
                if (
                    len(parts) >= 4
                    and parts[0].startswith("0x")
                    and parts[2] == "="
                    and parts[1] != "."
                ):
                    add_symbol(parts[1], parts[0])

    rom_start = symbols["_sfixed"][0]
    ram_start = symbols["_srelocate"][0]
    ram_end = symbols["_estack"][0]
    ram_length = ram_end - ram_start
    # print(ram_length, "ram length")
    # print(len(ram_dump) // ram_length, "snapshots")
    if analyze_snapshots == "all":
        snapshots = range(len(ram_dump) // ram_length - 1, -1, -1)
        # snapshots = range(4576, -1, -1)
    elif analyze_snapshots == "last":
        snapshots = range(len(ram_dump) // ram_length - 1, len(ram_dump) // ram_length - 2, -1)
    for snapshot_num in snapshots:
        ram = ram_dump[ram_length * snapshot_num : ram_length * (snapshot_num + 1)]

        ownership_graph = pgv.AGraph(directed=True)

        def load(address, size=4):
            if size is None:
                raise ValueError("You must provide a size")
            if address > ram_start:
                ram_address = address - ram_start
                if (ram_address + size) > len(ram):
                    raise ValueError("Unable to read 0x{:08x} from ram.".format(address))
                return ram[ram_address : ram_address + size]
            elif address < len(rom):
                if (address + size) > len(rom):
                    raise ValueError("Unable to read 0x{:08x} from rom.".format(address))
                return rom[address : address + size]

        def load_pointer(address):
            return struct.unpack("<I", load(address))[0]

        if "heap" in symbols:
            heap_start, heap_size = symbols["heap"]
        else:
            print("no static heap")
            allocations_start, allocations_size = symbols["allocations"]
            allocations = load(allocations_start, allocations_size)
            first_zero = True
            potential_heap = None
            # The heap is the last left hand allocated section that should span all the way to the
            # right side list.
            for address, size in struct.iter_unpack("<II", allocations):
                print(address, size)
                if address == 0 and first_zero:
                    first_zero = False
                if first_zero:
                    potential_heap = (address, size)

                if not first_zero and address != 0:
                    if address != potential_heap[0] + potential_heap[1]:
                        print("no active heap")
                        return
                    else:
                        heap_start, heap_size = potential_heap
                        break
        print("found heap", heap_start, heap_size)
        heap = load(heap_start, heap_size)
        total_byte_len = len(heap)

        # These change every run so we load them from the symbol table
        mp_state_ctx = symbols["mp_state_ctx"][0]
        manual_symbol_map["mp_state_ctx+20"] = "mp_state_ctx.vm.last_pool"
        last_pool = load_pointer(mp_state_ctx + 20)  # (gdb) p &mp_state_ctx.vm.last_pool
        manual_symbol_map["mp_state_ctx+108"] = "mp_state_ctx.vm.dict_main.map.table"
        dict_main_table = load_pointer(
            mp_state_ctx + 108
        )  # (gdb) p &mp_state_ctx.vm.dict_main.map.table
        manual_symbol_map["mp_state_ctx+84"] = "mp_state_ctx.vm.mp_loaded_modules_dict.map.table"
        imports_table = load_pointer(
            mp_state_ctx + 84
        )  # (gdb) p &mp_state_ctx.vm.mp_loaded_modules_dict.map.table

        manual_symbol_map["mp_state_ctx+124"] = "mp_state_ctx.vm.mp_sys_path_obj.items"
        manual_symbol_map["mp_state_ctx+140"] = "mp_state_ctx.vm.mp_sys_argv_obj.items"
        manual_symbol_map["mp_state_ctx+96"] = "mp_state_ctx.vm.dict_main"
        manual_symbol_map["0x200015e0"] = "mp_state_ctx.vm.dict_main"

        for i in range(READLINE_HIST_SIZE):
            manual_symbol_map[
                "mp_state_ctx+{}".format(148 + i * 4)
            ] = "mp_state_ctx.vm.readline_hist[{}]".format(i)

        tuple_type = symbols["mp_type_tuple"][0]
        type_type = symbols["mp_type_type"][0]
        map_type = symbols["mp_type_map"][0]
        dict_type = symbols["mp_type_dict"][0]
        property_type = symbols["mp_type_property"][0]
        str_type = symbols["mp_type_str"][0]
        function_types = [
            symbols["mp_type_fun_" + x][0]
            for x in ["bc", "builtin_0", "builtin_1", "builtin_2", "builtin_3", "builtin_var"]
        ]
        bytearray_type = symbols["mp_type_bytearray"][0]

        dynamic_type = 0x40000000  # placeholder, doesn't match any memory

        long_lived_start = load_pointer(
            mp_state_ctx + 272
        )  # (gdb) p &mp_state_ctx.mem.gc_lowest_long_lived_ptr

        type_colors = {
            dict_type: "red",
            property_type: "yellow",
            map_type: "blue",
            type_type: "orange",
            tuple_type: "skyblue",
            str_type: "pink",
            bytearray_type: "purple",
        }

        pool_shift = heap_start % BYTES_PER_BLOCK
        atb_length = (
            total_byte_len
            * BITS_PER_BYTE
            // (
                BITS_PER_BYTE
                + BITS_PER_BYTE * BLOCKS_PER_ATB // BLOCKS_PER_FTB
                + BITS_PER_BYTE * BLOCKS_PER_ATB * BYTES_PER_BLOCK
            )
        )
        pool_length = atb_length * BLOCKS_PER_ATB * BYTES_PER_BLOCK
        gc_finaliser_table_byte_len = (
            atb_length * BLOCKS_PER_ATB + BLOCKS_PER_FTB - 1
        ) // BLOCKS_PER_FTB

        if print_heap_structure:
            print("mp_state_ctx at 0x{:08x} and length {}".format(*symbols["mp_state_ctx"]))
            print("Total heap length:", total_byte_len)
            print("ATB length:", atb_length)
            print("Total allocatable:", pool_length)
            print("FTB length:", gc_finaliser_table_byte_len)

        pool_start = heap_start + total_byte_len - pool_length - pool_shift
        pool = heap[-pool_length - pool_shift :]

        total_height = 128 * 18
        total_width = (pool_length // (128 * 16)) * 85

        map_element_blocks = [dict_main_table, imports_table]
        string_blocks = []
        bytecode_blocks = []
        qstr_pools = []
        qstr_chunks = []
        block_data = {}

        # Find all the qtr pool addresses.
        prev_pool = last_pool
        while prev_pool > ram_start:
            qstr_pools.append(prev_pool)
            prev_pool = load_pointer(prev_pool)

        def save_allocated_block(end, current_allocation):
            allocation_length = current_allocation * BYTES_PER_BLOCK
            start = end - allocation_length
            address = pool_start + start
            data = pool[start:end]
            if print_block_state:
                print("0x{:x} {} bytes allocated".format(address, allocation_length))
            if print_block_contents:
                print(data)

            rows = ""
            for k in range(current_allocation - 1):
                rows += "<tr>"
                for l in range(4):
                    rows += '<td port="{}" height="18" width="20"></td>'.format(4 * (k + 1) + l)
                rows += "</tr>"
            table = '<<table bgcolor="gray" border="1" cellpadding="0" cellspacing="0"><tr><td colspan="4" port="0" height="18" width="80">0x{:08x}</td></tr>{}</table>>'.format(
                address, rows
            )

            ownership_graph.add_node(address, label=table, style="invisible", shape="plaintext")
            print("add  0x{:08x}".format(address))
            potential_type = None
            node = ownership_graph.get_node(address)
            node.attr["height"] = 0.25 * current_allocation
            if address >= long_lived_start:
                node.attr["fontcolor"] = "hotpink"
            else:
                node.attr["fontcolor"] = "black"
            block_data[address] = data
            for k in range(len(data) // 4):
                word = struct.unpack_from("<I", data, offset=(k * 4))[0]
                if word < len(rom) and k == 0 or address in qstr_pools:
                    potential_type = word
                    bgcolor = "gray"
                    if address in qstr_pools:
                        # print(address, len(data))
                        bgcolor = "tomato"
                    elif potential_type in function_types:
                        bgcolor = "green"
                    elif potential_type in type_colors:
                        bgcolor = type_colors[potential_type]
                    elif print_unknown_types:
                        print("unknown type", hex(potential_type))

                    node.attr["label"] = (
                        "<" + node.attr["label"].replace('"gray"', '"' + bgcolor + '"') + ">"
                    )

                if potential_type == str_type and k == 3:
                    string_blocks.append(word)

                if potential_type == dict_type:
                    if k == 3:
                        map_element_blocks.append(word)

                if ram_start < word < (ram_start + len(ram)) and word % 16 == 0:
                    port = k
                    if k < 4:
                        port = 0
                    ownership_graph.add_edge(address, word, tailport=str(port) + ":_")
                    print("  0x{:08x}".format(word))
                    if address in qstr_pools:
                        if k > 0:
                            qstr_chunks.append(word)
                    if k == 0:
                        potential_type = dynamic_type

                if potential_type == dynamic_type:
                    if k == 0:
                        node.attr["fillcolor"] = "plum"
                    if k == 3 and ram_start < word < ram_end:
                        map_element_blocks.append(word)

                if potential_type in function_types:
                    if k == 2 and ram_start < word < ram_end:
                        bytecode_blocks.append(word)

        longest_free = 0
        current_free = 0
        current_allocation = 0
        total_free = 0
        for i in range(atb_length):
            # Each atb byte is four blocks worth of info
            atb = heap[i]
            for j in range(4):
                block_state = (atb >> (j * 2)) & 0x3
                if block_state != AT_FREE and current_free > 0:
                    if print_block_state:
                        print("{} bytes free".format(current_free * BYTES_PER_BLOCK))
                    current_free = 0
                if block_state != AT_TAIL and current_allocation > 0:
                    save_allocated_block(
                        (i * BLOCKS_PER_ATB + j) * BYTES_PER_BLOCK, current_allocation
                    )
                    current_allocation = 0
                if block_state == AT_FREE:
                    current_free += 1
                    total_free += 1
                elif block_state == AT_HEAD or block_state == AT_MARK:
                    current_allocation = 1
                elif block_state == AT_TAIL and current_allocation > 0:
                    # In gc_free the logging happens before the tail is freed. So checking
                    # current_allocation > 0 ensures we only extend an allocation thats started.
                    current_allocation += 1
                longest_free = max(longest_free, current_free)
        # if current_free > 0:
        #    print("{} bytes free".format(current_free * BYTES_PER_BLOCK))
        if current_allocation > 0:
            save_allocated_block(pool_length, current_allocation)

        def is_qstr(obj):
            return obj & 0xFF800007 == 0x00000006

        def find_qstr(qstr_index):
            pool_ptr = last_pool
            if not is_qstr(qstr_index):
                return "object"

            pool = block_data[pool_ptr]
            prev, total_prev_len, alloc, length = struct.unpack_from("<IIII", pool)
            qstr_index >>= 3
            if qstr_index > total_prev_len + alloc:
                return "invalid"
            while pool_ptr != 0:
                if pool_ptr > ram_start:
                    if pool_ptr in block_data:
                        pool = block_data[pool_ptr]
                        prev, total_prev_len, alloc, length = struct.unpack_from("<IIII", pool)
                    else:
                        print("missing qstr pool: {:08x}".format(pool_ptr))
                        return "missing"
                else:
                    rom_offset = pool_ptr - rom_start
                    prev, total_prev_len, alloc, length = struct.unpack_from(
                        "<IIII", rom[rom_offset : rom_offset + 32]
                    )
                    pool = rom[rom_offset : rom_offset + 32 + length * 4]

                if qstr_index >= total_prev_len:
                    offset = (qstr_index - total_prev_len) * 4 + 16
                    start = struct.unpack_from("<I", pool, offset=offset)[0]
                    if start < heap_start:
                        start -= rom_start
                        if start > len(rom):
                            return "more than rom: {:x}".format(start + rom_start)
                        qstr_hash, qstr_len = struct.unpack("<BB", rom[start : start + 2])
                        return rom[start + 2 : start + 2 + qstr_len].decode("utf-8")
                    else:
                        if start > heap_start + len(heap):
                            return "out of range: {:x}".format(start)
                        local = start - heap_start
                        qstr_hash, qstr_len = struct.unpack("<BB", heap[local : local + 2])
                        return heap[local + 2 : local + 2 + qstr_len].decode("utf-8")

                pool_ptr = prev
            return "unknown"

        def format(obj):
            if obj & 1 != 0:
                return obj >> 1
            if is_qstr(obj):
                return find_qstr(obj)
            else:
                return "0x{:08x}".format(obj)

        for block in sorted(map_element_blocks):
            if block == 0:
                continue
            try:
                node = ownership_graph.get_node(block)
            except KeyError:
                print(
                    "Unable to find memory block for 0x{:08x}. Is there something running?".format(
                        block
                    )
                )
                continue
            if block not in block_data:
                continue
            data = block_data[block]
            cells = []
            for i in range(len(data) // 8):
                key, value = struct.unpack_from("<II", data, offset=(i * 8))
                if key == MP_OBJ_NULL or key == MP_OBJ_SENTINEL:
                    cells.append(("", " "))
                else:
                    cells.append((key, format(key)))
                    if value in block_data:
                        edge = ownership_graph.get_edge(block, value)
                        edge.attr["tailport"] = str(key)
            rows = ""
            for i in range(len(cells) // 2):
                rows += '<tr><td port="{}">{}</td><td port="{}">{}</td></tr>'.format(
                    cells[2 * i][0], cells[2 * i][1], cells[2 * i + 1][0], cells[2 * i + 1][1]
                )
            node.attr["shape"] = "plaintext"
            node.attr["style"] = "invisible"
            node.attr[
                "label"
            ] = '<<table bgcolor="gold" border="1" cellpadding="0" cellspacing="0"><tr><td colspan="2">0x{:08x}</td></tr>{}</table>>'.format(
                block, rows
            )

        for node, degree in ownership_graph.in_degree_iter():
            print(node, degree)
            if degree == 0:
                address_bytes = struct.pack("<I", int(node))
                location = -1
                for _ in range(ram.count(address_bytes)):
                    location = ram.find(address_bytes, location + 1)
                    pointer_location = ram_start + location
                    source = "0x{:08x}".format(pointer_location)
                    if pointer_location in symbol_lookup:
                        source = symbol_lookup[pointer_location]
                    if source in manual_symbol_map:
                        source = manual_symbol_map[source]
                    if "readline_hist" in source:
                        string_blocks.append(int(node))
                    if pointer_location > heap_start + heap_size:
                        source = "stack " + source
                    ownership_graph.add_edge(source, node)

        for block in string_blocks:
            if block == 0:
                continue
            node = ownership_graph.get_node(block)
            node.attr["fillcolor"] = "hotpink"
            if block in block_data:
                raw_string = block_data[block]
            else:
                print("Unable to find memory block for string at 0x{:08x}.".format(block))
                continue
            try:
                raw_string = block_data[block].decode("utf-8")
            except:
                raw_string = str(block_data[block])
            wrapped = []
            for i in range(0, len(raw_string), 16):
                wrapped.append(raw_string[i : i + 16])
            node.attr["label"] = "\n".join(wrapped)
            node.attr["style"] = "filled"
            node.attr["fontname"] = "FiraCode-Medium"
            node.attr["fontpath"] = "/Users/tannewt/Library/Fonts/"
            node.attr["fontsize"] = 8
            node.attr["height"] = len(wrapped) * 0.25

        for block in bytecode_blocks:
            node = ownership_graph.get_node(block)
            node.attr["fillcolor"] = "lightseagreen"
            if block in block_data:
                data = block_data[block]
            else:
                print("Unable to find memory block for bytecode at 0x{:08x}.".format(block))
                continue
            prelude = Prelude(io.BufferedReader(io.BytesIO(data)))
            node.attr["shape"] = "plaintext"
            node.attr["style"] = "invisible"
            code_info_size = prelude.code_info_size
            rows = ""
            remaining_bytecode = len(data) - 16
            while code_info_size >= 16:
                rows += (
                    '<tr><td colspan="16" bgcolor="palegreen" height="18" width="80"></td></tr>'
                )
                code_info_size -= 16
                remaining_bytecode -= 16
            if code_info_size > 0:
                rows += (
                    '<tr><td colspan="{}" bgcolor="palegreen" height="18" width="{}"></td>'
                    '<td colspan="{}" bgcolor="seagreen" height="18" width="{}"></td></tr>'
                ).format(
                    code_info_size,
                    code_info_size * (80 / 16),
                    (16 - code_info_size),
                    (80 / 16) * (16 - code_info_size),
                )
                remaining_bytecode -= 16
            for i in range(remaining_bytecode // 16):
                rows += '<tr><td colspan="16" bgcolor="seagreen" height="18" width="80"></td></tr>'
            node.attr[
                "label"
            ] = '<<table border="1" cellspacing="0"><tr><td colspan="16" bgcolor="lightseagreen" height="18" width="80">0x{:08x}</td></tr>{}</table>>'.format(
                block, rows
            )

        for block in qstr_chunks:
            if block not in block_data:
                ownership_graph.delete_node(block)
                continue
            data = block_data[block]
            qstrs_in_chunk = ""
            offset = 0
            while offset < len(data) - 1:
                qstr_hash, qstr_len = struct.unpack_from("<BB", data, offset=offset)
                if qstr_hash == 0:
                    qstrs_in_chunk += " " * (len(data) - offset)
                    offset = len(data)
                    continue
                offset += 2 + qstr_len + 1
                try:
                    qstrs_in_chunk += "  " + data[offset - qstr_len - 1 : offset - 1].decode(
                        "utf-8"
                    )
                except UnicodeDecodeError:
                    qstrs_in_chunk += "  " + "░" * qstr_len
            printable_qstrs = ""
            for i in range(len(qstrs_in_chunk)):
                c = qstrs_in_chunk[i]
                if c not in string.printable or c in "\v\f":
                    printable_qstrs += "░"
                else:
                    printable_qstrs += qstrs_in_chunk[i]
            wrapped = []
            for i in range(0, len(printable_qstrs), 16):
                wrapped.append(html.escape(printable_qstrs[i : i + 16]))
            node = ownership_graph.get_node(block)
            node.attr[
                "label"
            ] = '<<table border="1" cellspacing="0" bgcolor="lightsalmon" width="80"><tr><td height="18" >0x{:08x}</td></tr><tr><td height="{}" >{}</td></tr></table>>'.format(
                block, 18 * (len(wrapped) - 1), "<br/>".join(wrapped)
            )
            node.attr["fontname"] = "FiraCode-Bold"
            if block >= long_lived_start:
                node.attr["fontcolor"] = "hotpink"
            else:
                node.attr["fontcolor"] = "black"
            node.attr["fontpath"] = "/Users/tannewt/Library/Fonts/"
            node.attr["fontsize"] = 8

        print("Total free space:", BYTES_PER_BLOCK * total_free)
        print("Longest free space:", BYTES_PER_BLOCK * longest_free)

        # First render the graph of objects on the heap.
        if draw_heap_ownership:
            ownership_graph.layout(prog="dot")
            fn = os.path.join(output_directory, "heap_ownership{:04d}.svg".format(snapshot_num))
            print(fn)
            ownership_graph.draw(fn)

        # Clear edge positioning from ownership graph layout.
        if draw_heap_ownership:
            for edge in ownership_graph.iteredges():
                del edge.attr["pos"]
        else:
            for edge in ownership_graph.edges():
                ownership_graph.delete_edge(edge)

        # Second, render the heap layout in memory order.
        for node in ownership_graph.nodes():
            try:
                address = int(node.name)
            except ValueError:
                ownership_graph.remove_node(node.name)
                continue
            block = (address - pool_start) // 16
            x = block // 128
            y = 128 - block % 128
            try:
                height = float(node.attr["height"])
            except:
                height = 0.25
            # print(hex(address), "height", height, y)
            # if address in block_data:
            #    print(hex(address), block, len(block_data[address]), x, y, height)
            node.attr["pos"] = "{},{}".format(x * 80, (y - (height - 0.25) * 2) * 18)  # in inches

        # Reformat block nodes so they are the correct size and do not have keys in them.
        for block in sorted(map_element_blocks):
            try:
                node = ownership_graph.get_node(block)
            except KeyError:
                if block != 0:
                    print(
                        "Unable to find memory block for 0x{:08x}. Is there something running?".format(
                            block
                        )
                    )
                continue
            # node.attr["fillcolor"] = "gold"
            if block not in block_data:
                continue
            data = block_data[block]
            # print("0x{:08x}".format(block))
            cells = []
            for i in range(len(data) // 8):
                key, value = struct.unpack_from("<II", data, offset=(i * 8))
                if key == MP_OBJ_NULL or key == MP_OBJ_SENTINEL:
                    # print("  <empty slot>")
                    cells.append(("", " "))
                else:
                    # print("  {}, {}".format(format(key), format(value)))
                    cells.append((key, ""))
                    # if value in block_data:
                    #     edge = ownership_graph.get_edge(block, value)
                    #     edge.attr["tailport"] = str(key)
            rows = ""
            for i in range(len(cells) // 2):
                rows += '<tr><td port="{}" height="18" width="40">{}</td><td port="{}" height="18" width="40">{}</td></tr>'.format(
                    cells[2 * i][0], cells[2 * i][1], cells[2 * i + 1][0], cells[2 * i + 1][1]
                )
            node.attr[
                "label"
            ] = '<<table bgcolor="gold" border="1" cellpadding="0" cellspacing="0">{}</table>>'.format(
                rows
            )

        ownership_graph.add_node(
            "center",
            pos="{},{}".format(total_width // 2 - 40, total_height // 2),
            shape="plaintext",
            label=" ",
        )
        ownership_graph.graph_attr["viewport"] = "{},{},1,{}".format(
            total_width, total_height, "center"
        )

        ownership_graph.has_layout = True

        if draw_heap_layout:
            fn = os.path.join(output_directory, "heap_layout{:04d}.png".format(snapshot_num))
            print(fn)
            # ownership_graph.write(fn+".dot")
            ownership_graph.draw(fn)


if __name__ == "__main__":
    do_all_the_things()



class Base:

    def __init__(self):
        self.a = 1

    def meth(self):
        print("in Base meth", self.a)

class Sub(Base):

    def meth(self):
        print("in Sub meth")
        return super().meth()

a = Sub()
a.meth()

# printing super
class A:
    def p(self):
        print(str(super())[:18])
A().p()


# test compiler's handling of long expressions with super
class A:
    bar = 123
    def foo(self):
        print('A foo')
        return [1, 2, 3]
class B(A):
    def foo(self):
        print('B foo')
        print(super().bar) # accessing attribute after super()
        return super().foo().count(2) # calling a subsequent method
print(B().foo())

try:
    super(1, 1).x
except TypeError:
    print(True)
else:
    print(False)

# first arg to super must be a type
try:
    super(1, 1)
except TypeError:
    print('TypeError')

# store/delete of super attribute not allowed
assert hasattr(super(B, B()), 'foo')
try:
    super(B, B()).foo = 1
except AttributeError:
    print('AttributeError')
try:
    del super(B, B()).foo
except AttributeError:
    print('AttributeError')



# SPDX-FileCopyrightText: 2014 MicroPython & CircuitPython contributors (https://github.com/adafruit/circuitpython/graphs/contributors)
#
# SPDX-License-Identifier: MIT
import os
import site
from datetime import datetime
from typing import Dict, List

from setuptools import setup
from pathlib import Path

STD_PACKAGES = set(('array', 'math', 'os', 'random', 'struct', 'sys', 'ssl', 'time'))

stub_root = Path("circuitpython-stubs")
stubs = [p.relative_to(stub_root).as_posix() for p in stub_root.glob("*.pyi")]

def local_scheme(version):
    return ""

packages = set(os.listdir("circuitpython-stubs")) - STD_PACKAGES
package_dir = dict((f"{package}-stubs", f"circuitpython-stubs/{package}")
    for package in packages)
print("package dir is", package_dir)

def build_package_data() -> Dict[str, List[str]]:
    result = {}
    for package in packages:
        result[f"{package}-stubs"] = ["*.pyi", "*/*.pyi"]
    return result

package_data=build_package_data()
setup(
    name="circuitpython-stubs",
    description="PEP 561 type stubs for CircuitPython",
    url="https://github.com/adafruit/circuitpython",
    maintainer="CircuitPythonistas",
    maintainer_email="circuitpython@adafruit.com",
    author_email="circuitpython@adafruit.com",
    license="MIT",
    packages=list(package_data.keys()),
    package_data=package_data,
    package_dir = package_dir,
    setup_requires=["setuptools_scm", "setuptools>=38.6.0"],
    use_scm_version={"local_scheme": local_scheme},
    zip_safe=False,
)



# test with handling within a native function


class C:
    def __init__(self):
        print("__init__")

    def __enter__(self):
        print("__enter__")

    def __exit__(self, a, b, c):
        print("__exit__", a, b, c)


# basic with
@micropython.native
def f():
    with C():
        print(1)


f()

# nested with and try-except
@micropython.native
def f():
    try:
        with C():
            print(1)
            fail
            print(2)
    except NameError:
        print("NameError")


f()



import ujson as json

# Test that json can load from any object with readinto


class Buffer:
    def __init__(self, data):
        self._data = data
        self._i = 0

    def readinto(self, buf):
        end = self._i + len(buf)
        remaining = len(self._data) - self._i
        end = min(end, len(self._data))
        l = min(len(buf), remaining)
        buf[:l] = self._data[self._i : end]
        self._i += l
        return l


print(json.load(Buffer(b"null")))
print(json.load(Buffer(b'"abc\\u0064e"')))
print(json.load(Buffer(b"[false, true, 1, -2]")))
print(json.load(Buffer(b'{"a":true}')))



class A:

    var = 132

    def __init__(self):
        self.var2 = 34

    def meth(self, i):
        return 42 + i


a = A()
print(hasattr(a, "var"))
print(hasattr(a, "var2"))
print(hasattr(a, "meth"))
print(hasattr(a, "_none_such"))
print(hasattr(list, "foo"))

class C:

    def __getattr__(self, attr):
        if attr == "exists":
            return attr
        elif attr == "raise":
            raise Exception(123)
        raise AttributeError

c = C()
print(hasattr(c, "exists"))
print(hasattr(c, "doesnt_exist"))

# ensure that non-AttributeError exceptions propagate out of hasattr
try:
    hasattr(c, "raise")
except Exception as er:
    print(er)

try:
    hasattr(1, b'123')
except TypeError:
    print('TypeError')

try:
    hasattr(1, 123)
except TypeError:
    print('TypeError')



# test class with __add__ and __sub__ methods

class C:
    def __init__(self, value):
        self.value = value

    def __add__(self, rhs):
        print(self.value, '+', rhs)

    def __sub__(self, rhs):
        print(self.value, '-', rhs)

c = C(0)
c + 1
c - 2



try:
    try:
        import ubinascii as binascii
    except ImportError:
        import binascii
except ImportError:
    print("SKIP")
    raise SystemExit

print(binascii.hexlify(b"\x00\x01\x02\x03\x04\x05\x06\x07"))
print(binascii.hexlify(b"\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"))
print(binascii.hexlify(b"\x7f\x80\xff"))
print(binascii.hexlify(b"1234ABCDabcd"))
try:
    binascii.hexlify("")
except TypeError:
    print("TypeError")



import audiocore
import audiobusio
import board
import digitalio
import array
import time
import math
import rp2pio
import adafruit_pioasm

time.sleep(10)

trigger = digitalio.DigitalInOut(board.D4)
trigger.switch_to_output(True)

# Generate one period of sine wav.
length = 8000 // 440

# signed 16 bit
s16 = array.array("h", [0] * length)
for i in range(length):
    s16[i] = int(math.sin(math.pi * 2 * i / length) * (2 ** 15))
    print(s16[i])

sample = audiocore.RawSample(s16, sample_rate=8000)

dac = audiobusio.I2SOut(bit_clock=board.D10, word_select=board.D11, data=board.D12)

trigger.value = False
dac.play(sample, loop=True)
time.sleep(1)
dac.stop()
trigger.value = True

print("done")



def gen():
    try:
        yield 1
    except ValueError as e:
        print("got ValueError from upstream!", repr(e.args))
    yield "str1"
    raise TypeError

def gen2():
    print((yield from gen()))

g = gen2()
print(next(g))
print(g.throw(ValueError))
try:
    print(next(g))
except TypeError:
    print("got TypeError from downstream!")

# passing None as second argument to throw
g = gen2()
print(next(g))
print(g.throw(ValueError, None))
try:
    print(next(g))
except TypeError:
    print("got TypeError from downstream!")

# passing an exception instance as second argument to throw
g = gen2()
print(next(g))
print(g.throw(ValueError, ValueError(123)))
try:
    print(next(g))
except TypeError:
    print("got TypeError from downstream!")

# thrown value is caught and then generator returns normally
def gen():
    try:
        yield 123
    except ValueError:
        print('ValueError')
    # return normally after catching thrown exception
def gen2():
    yield from gen()
    yield 789
g = gen2()
print(next(g))
print(g.throw(ValueError))



# overriding default arguments

def foo(a, b=3):
    print(a, b)

# override with positional
foo(1, 333)

# override with keyword
foo(1, b=333)

# override with keyword
foo(a=2, b=333)

def foo2(a=1, b=2):
    print(a, b)

# default and keyword
foo2(b='two')



#!/usr/bin/env python3

# Microsoft UF2
#
# The MIT License (MIT)
#
# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import struct
import subprocess
import re
import os
import os.path
import argparse


UF2_MAGIC_START0 = 0x0A324655  # "UF2\n"
UF2_MAGIC_START1 = 0x9E5D5157  # Randomly selected
UF2_MAGIC_END = 0x0AB16F30  # Ditto

families = {
    "SAMD21": 0x68ED2B88,
    "SAMD51": 0x55114460,
    "NRF52": 0x1B57745F,
    "STM32F1": 0x5EE21072,
    "STM32F4": 0x57755A57,
    "ATMEGA32": 0x16573617,
}

INFO_FILE = "/INFO_UF2.TXT"

appstartaddr = 0x2000
familyid = 0x0


def is_uf2(buf):
    w = struct.unpack("<II", buf[0:8])
    return w[0] == UF2_MAGIC_START0 and w[1] == UF2_MAGIC_START1


def is_hex(buf):
    try:
        w = buf[0:30].decode("utf-8")
    except UnicodeDecodeError:
        return False
    if w[0] == ":" and re.match(b"^[:0-9a-fA-F\r\n]+$", buf):
        return True
    return False


def convert_from_uf2(buf):
    global appstartaddr
    numblocks = len(buf) // 512
    curraddr = None
    outp = b""
    for blockno in range(numblocks):
        ptr = blockno * 512
        block = buf[ptr : ptr + 512]
        hd = struct.unpack(b"<IIIIIIII", block[0:32])
        if hd[0] != UF2_MAGIC_START0 or hd[1] != UF2_MAGIC_START1:
            print("Skipping block at " + ptr + "; bad magic")
            continue
        if hd[2] & 1:
            # NO-flash flag set; skip block
            continue
        datalen = hd[4]
        if datalen > 476:
            assert False, "Invalid UF2 data size at " + ptr
        newaddr = hd[3]
        if curraddr == None:
            appstartaddr = newaddr
            curraddr = newaddr
        padding = newaddr - curraddr
        if padding < 0:
            assert False, "Block out of order at " + ptr
        if padding > 10 * 1024 * 1024:
            assert False, "More than 10M of padding needed at " + ptr
        if padding % 4 != 0:
            assert False, "Non-word padding size at " + ptr
        while padding > 0:
            padding -= 4
            outp += b"\x00\x00\x00\x00"
        outp += block[32 : 32 + datalen]
        curraddr = newaddr + datalen
    return outp


def convert_to_carray(file_content):
    outp = "const unsigned char bindata[] __attribute__((aligned(16))) = {"
    for i in range(len(file_content)):
        if i % 16 == 0:
            outp += "\n"
        outp += "0x%02x, " % ord(file_content[i])
    outp += "\n};\n"
    return outp


def convert_to_uf2(file_content):
    global familyid
    datapadding = b""
    while len(datapadding) < 512 - 256 - 32 - 4:
        datapadding += b"\x00\x00\x00\x00"
    numblocks = (len(file_content) + 255) // 256
    outp = b""
    for blockno in range(numblocks):
        ptr = 256 * blockno
        chunk = file_content[ptr : ptr + 256]
        flags = 0x0
        if familyid:
            flags |= 0x2000
        hd = struct.pack(
            b"<IIIIIIII",
            UF2_MAGIC_START0,
            UF2_MAGIC_START1,
            flags,
            ptr + appstartaddr,
            256,
            blockno,
            numblocks,
            familyid,
        )
        while len(chunk) < 256:
            chunk += b"\x00"
        block = hd + chunk + datapadding + struct.pack(b"<I", UF2_MAGIC_END)
        assert len(block) == 512
        outp += block
    return outp


class Block:
    def __init__(self, addr):
        self.addr = addr
        self.bytes = bytearray(256)

    def encode(self, blockno, numblocks):
        global familyid
        flags = 0x0
        if familyid:
            flags |= 0x2000
        hd = struct.pack(
            "<IIIIIIII",
            UF2_MAGIC_START0,
            UF2_MAGIC_START1,
            flags,
            self.addr,
            256,
            blockno,
            numblocks,
            familyid,
        )
        hd += self.bytes[0:256]
        while len(hd) < 512 - 4:
            hd += b"\x00"
        hd += struct.pack("<I", UF2_MAGIC_END)
        return hd


def convert_from_hex_to_uf2(buf):
    global appstartaddr
    appstartaddr = None
    upper = 0
    currblock = None
    blocks = []
    for line in buf.split("\n"):
        if line[0] != ":":
            continue
        i = 1
        rec = []
        while i < len(line) - 1:
            rec.append(int(line[i : i + 2], 16))
            i += 2
        tp = rec[3]
        if tp == 4:
            upper = ((rec[4] << 8) | rec[5]) << 16
        elif tp == 2:
            upper = ((rec[4] << 8) | rec[5]) << 4
            assert (upper & 0xFFFF) == 0
        elif tp == 1:
            break
        elif tp == 0:
            addr = upper | (rec[1] << 8) | rec[2]
            if appstartaddr == None:
                appstartaddr = addr
            i = 4
            while i < len(rec) - 1:
                if not currblock or currblock.addr & ~0xFF != addr & ~0xFF:
                    currblock = Block(addr & ~0xFF)
                    blocks.append(currblock)
                currblock.bytes[addr & 0xFF] = rec[i]
                addr += 1
                i += 1
    numblocks = len(blocks)
    resfile = b""
    for i in range(0, numblocks):
        resfile += blocks[i].encode(i, numblocks)
    return resfile


def get_drives():
    drives = []
    if sys.platform == "win32":
        r = subprocess.check_output(
            [
                "wmic",
                "PATH",
                "Win32_LogicalDisk",
                "get",
                "DeviceID,",
                "VolumeName,",
                "FileSystem,",
                "DriveType",
            ]
        )
        for line in r.split("\n"):
            words = re.split("\s+", line)
            if len(words) >= 3 and words[1] == "2" and words[2] == "FAT":
                drives.append(words[0])
    else:
        rootpath = "/media"
        if sys.platform == "darwin":
            rootpath = "/Volumes"
        elif sys.platform == "linux":
            tmp = rootpath + "/" + os.environ["USER"]
            if os.path.isdir(tmp):
                rootpath = tmp
        for d in os.listdir(rootpath):
            drives.append(os.path.join(rootpath, d))

    def has_info(d):
        try:
            return os.path.isfile(d + INFO_FILE)
        except:
            return False

    return list(filter(has_info, drives))


def board_id(path):
    with open(path + INFO_FILE, mode="r") as file:
        file_content = file.read()
    return re.search("Board-ID: ([^\r\n]*)", file_content).group(1)


def list_drives():
    for d in get_drives():
        print(d, board_id(d))


def write_file(name, buf):
    with open(name, "wb") as f:
        f.write(buf)
    print("Wrote %d bytes to %s." % (len(buf), name))


def main():
    global appstartaddr, familyid

    def error(msg):
        print(msg)
        sys.exit(1)

    parser = argparse.ArgumentParser(description="Convert to UF2 or flash directly.")
    parser.add_argument(
        "input", metavar="INPUT", type=str, nargs="?", help="input file (HEX, BIN or UF2)"
    )
    parser.add_argument(
        "-b",
        "--base",
        dest="base",
        type=str,
        default="0x2000",
        help="set base address of application for BIN format (default: 0x2000)",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        dest="output",
        type=str,
        help='write output to named file; defaults to "flash.uf2" or "flash.bin" where sensible',
    )
    parser.add_argument("-d", "--device", dest="device_path", help="select a device path to flash")
    parser.add_argument("-l", "--list", action="store_true", help="list connected devices")
    parser.add_argument("-c", "--convert", action="store_true", help="do not flash, just convert")
    parser.add_argument(
        "-f",
        "--family",
        dest="family",
        type=str,
        default="0x0",
        help="specify familyID - number or name (default: 0x0)",
    )
    parser.add_argument(
        "-C", "--carray", action="store_true", help="convert binary file to a C array, not UF2"
    )
    args = parser.parse_args()
    appstartaddr = int(args.base, 0)

    if args.family.upper() in families:
        familyid = families[args.family.upper()]
    else:
        try:
            familyid = int(args.family, 0)
        except ValueError:
            error("Family ID needs to be a number or one of: " + ", ".join(families.keys()))

    if args.list:
        list_drives()
    else:
        if not args.input:
            error("Need input file")
        with open(args.input, mode="rb") as f:
            inpbuf = f.read()
        from_uf2 = is_uf2(inpbuf)
        ext = "uf2"
        if from_uf2:
            outbuf = convert_from_uf2(inpbuf)
            ext = "bin"
        elif is_hex(inpbuf):
            outbuf = convert_from_hex_to_uf2(inpbuf.decode("utf-8"))
        elif args.carray:
            outbuf = convert_to_carray(inpbuf)
            ext = "h"
        else:
            outbuf = convert_to_uf2(inpbuf)
        print(
            "Converting to %s, output size: %d, start address: 0x%x"
            % (ext, len(outbuf), appstartaddr)
        )
        if args.convert:
            drives = []
            if args.output == None:
                args.output = "flash." + ext
        else:
            drives = get_drives()

        if args.output:
            write_file(args.output, outbuf)
        else:
            if len(drives) == 0:
                error("No drive to deploy.")
        for d in drives:
            print("Flashing %s (%s)" % (d, board_id(d)))
            write_file(d + "/NEW.UF2", outbuf)


if __name__ == "__main__":
    main()



# test _thread lock objects where a lock is acquired/released by a different thread

import _thread


def thread_entry():
    print("thread about to release lock")
    lock.release()


lock = _thread.allocate_lock()
lock.acquire()
_thread.start_new_thread(thread_entry, ())
lock.acquire()
print("main has lock")
lock.release()



try:
    import usocket as socket, uselect as select, uerrno as errno
except ImportError:
    try:
        import socket, select, errno

        select.poll  # Raises AttributeError for CPython implementations without poll()
    except (ImportError, AttributeError):
        print("SKIP")
        raise SystemExit


poller = select.poll()

s = socket.socket()

poller.register(s)
# https://docs.python.org/3/library/select.html#select.poll.register
# "Registering a file descriptor that’s already registered is not an error,
# and has the same effect as registering the descriptor exactly once."
poller.register(s)

# 2 args are mandatory unlike register()
try:
    poller.modify(s)
except TypeError:
    print("modify:TypeError")

poller.modify(s, select.POLLIN)

poller.unregister(s)

try:
    poller.modify(s, select.POLLIN)
except OSError as e:
    assert e.errno == errno.ENOENT

# poll after closing the socket, should return POLLNVAL
poller.register(s)
s.close()
p = poller.poll(0)
print(len(p), p[0][-1])



# Test for behaviour of combined standard and extended block device

try:
    import uos

    uos.VfsFat
    uos.VfsLfs2
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class RAMBlockDevice:
    ERASE_BLOCK_SIZE = 512

    def __init__(self, blocks):
        self.data = bytearray(blocks * self.ERASE_BLOCK_SIZE)

    def readblocks(self, block, buf, off=0):
        addr = block * self.ERASE_BLOCK_SIZE + off
        for i in range(len(buf)):
            buf[i] = self.data[addr + i]

    def writeblocks(self, block, buf, off=None):
        if off is None:
            # erase, then write
            off = 0
        addr = block * self.ERASE_BLOCK_SIZE + off
        for i in range(len(buf)):
            self.data[addr + i] = buf[i]

    def ioctl(self, op, arg):
        if op == 4:  # block count
            return len(self.data) // self.ERASE_BLOCK_SIZE
        if op == 5:  # block size
            return self.ERASE_BLOCK_SIZE
        if op == 6:  # erase block
            return 0


def test(bdev, vfs_class):
    print("test", vfs_class)

    # mkfs
    vfs_class.mkfs(bdev)

    # construction
    vfs = vfs_class(bdev)

    # statvfs
    print(vfs.statvfs("/"))

    # open, write close
    f = vfs.open("test", "w")
    for i in range(10):
        f.write("some data")
    f.close()

    # ilistdir
    print(list(vfs.ilistdir()))

    # read
    with vfs.open("test", "r") as f:
        print(f.read())


try:
    bdev = RAMBlockDevice(50)
except MemoryError:
    print("SKIP")
    raise SystemExit

test(bdev, uos.VfsFat)
test(bdev, uos.VfsLfs2)



#!/usr/bin/env python3

import sys
import re
import os
import itertools
from concurrent.futures import ProcessPoolExecutor
import multiprocessing
import threading
import subprocess

from makeqstrdefs import qstr_unescape, QSTRING_BLOCK_LIST

re_line = re.compile(r"#[line]*\s(\d+)\s\"([^\"]+)\"", re.DOTALL)
re_qstr = re.compile(r"MP_QSTR_[_a-zA-Z0-9]+", re.DOTALL)
re_translate = re.compile(r"translate\(\"((?:(?=(\\?))\2.)*?)\"\)", re.DOTALL)


def write_out(fname, output_dir, output):
    if output:
        for m, r in [("/", "__"), ("\\", "__"), (":", "@"), ("..", "@@")]:
            fname = fname.replace(m, r)
        with open(output_dir + "/" + fname + ".qstr", "w") as f:
            f.write("\n".join(output) + "\n")


def process_file(fname, output_dir, content):
    content = content.decode("utf-8", errors="ignore")
    output = []
    for match in re_qstr.findall(content):
        name = match.replace("MP_QSTR_", "")
        if name not in QSTRING_BLOCK_LIST:
            output.append("Q(" + qstr_unescape(name) + ")")
    for match in re_translate.findall(content):
        output.append('TRANSLATE("' + match[0] + '")')

    write_out(fname, output_dir, output)


def checkoutput1(args):
    info = subprocess.run(args, check=True, stdout=subprocess.PIPE, input="")
    return info.stdout


def preprocess(command, output_dir, fn):
    try:
        output = checkoutput1(command + [fn])
        process_file(fn, output_dir, output)
    except Exception as e:
        print(e, file=sys.stderr)


def maybe_preprocess(command, output_dir, fn):
    # Preprocess the source file if it contains "MP_QSTR", "translate",
    # or if it uses enum.h (which generates "MP_QSTR" strings.
    if subprocess.call(["grep", "-lqE", r"(MP_QSTR|translate|enum\.h)", fn]) == 0:
        preprocess(command, output_dir, fn)


if __name__ == "__main__":

    idx1 = sys.argv.index("--")
    idx2 = sys.argv.index("--", idx1 + 1)
    output_dir = sys.argv[1]
    check = sys.argv[2:idx1]
    always = sys.argv[idx1 + 1 : idx2]
    command = sys.argv[idx2 + 1 :]

    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    # Mac and Windows use 'spawn'.  Uncomment this during testing to catch spawn-specific problems on Linux.
    # multiprocessing.set_start_method("spawn")
    executor = ProcessPoolExecutor(max_workers=multiprocessing.cpu_count() + 1)
    executor.map(maybe_preprocess, itertools.repeat(command), itertools.repeat(output_dir), check)
    executor.map(preprocess, itertools.repeat(command), itertools.repeat(output_dir), always)
    executor.shutdown()



# test slices; only 2 argument version supported by MicroPython at the moment
x = list(range(10))

# Assignment
l = list(x)
l[1:3] = [10, 20]
print(l)
l = list(x)
l[1:3] = [10]
print(l)
l = list(x)
l[1:3] = []
print(l)
l = list(x)
del l[1:3]
print(l)

l = list(x)
l[:3] = [10, 20]
print(l)
l = list(x)
l[:3] = []
print(l)
l = list(x)
del l[:3]
print(l)

l = list(x)
l[:-3] = [10, 20]
print(l)
l = list(x)
l[:-3] = []
print(l)
l = list(x)
del l[:-3]
print(l)

# assign a tuple
l = [1, 2, 3]
l[0:1] = (10, 11, 12)
print(l)

# RHS of slice must be an iterable
try:
    [][0:1] = 123
except TypeError:
    print('TypeError')



# test scoping rules

# explicit global variable
a = 1
def f():
    global a
    global a, a # should be able to redefine as global
    a = 2
f()
print(a)

# explicit nonlocal variable
def f():
    a = 1
    def g():
        nonlocal a
        nonlocal a, a # should be able to redefine as nonlocal
        a = 2
    g()
    return a
print(f())

# nonlocal at inner-inner level (h)
def f():
    x = 1
    def g():
        def h():
            nonlocal x
            return x
        return h
    return g
print(f()()())

# nonlocal declared at outer level (g), and referenced by inner level (h)
def f():
    x = 1
    def g():
        nonlocal x
        def h():
            return x
        return h
    return g
print(f()()())



import pkg.mod

print(pkg.__name__)
print(pkg.mod.__name__)
print(pkg.mod.foo())

# Import 2nd time, must be same module objects
pkg_ = __import__("pkg.mod")
print(pkg_ is not pkg.mod)
print(pkg_ is pkg)
print(pkg_.mod is pkg.mod)

# import using "as"
import pkg.mod as mm

print(mm is pkg.mod)
print(mm.foo())



# test builtin "all" and "any"

tests = (
    (),
    [],
    [False],
    [True],
    [False, True],
    [True, False],
    [False, False],
    [True, True],
    range(10),
)

for test in tests:
    print(all(test))

for test in tests:
    print(any(test))



# test VFS functionality with a user-defined filesystem
# also tests parts of uio.IOBase implementation

import sys

try:
    import uio

    uio.IOBase
    import uos

    uos.mount
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class UserFile(uio.IOBase):
    def __init__(self, mode, data):
        assert isinstance(data, bytes)
        self.is_text = mode.find("b") == -1
        self.data = data
        self.pos = 0

    def read(self):
        if self.is_text:
            return str(self.data, "utf8")
        else:
            return self.data

    def readinto(self, buf):
        assert not self.is_text
        n = 0
        while n < len(buf) and self.pos < len(self.data):
            buf[n] = self.data[self.pos]
            n += 1
            self.pos += 1
        return n

    def ioctl(self, req, arg):
        print("ioctl", req, arg)
        return 0


class UserFS:
    def __init__(self, files):
        self.files = files

    def mount(self, readonly, mksfs):
        pass

    def umount(self):
        pass

    def stat(self, path):
        print("stat", path)
        if path in self.files:
            return (32768, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        raise OSError

    def open(self, path, mode):
        print("open", path, mode)
        return UserFile(mode, self.files[path])


# create and mount a user filesystem
user_files = {
    "/data.txt": b"some data in a text file",
    "/usermod1.py": b"print('in usermod1')\nimport usermod2",
    "/usermod2.py": b"print('in usermod2')",
}
uos.mount(UserFS(user_files), "/userfs")

# open and read a file
f = open("/userfs/data.txt")
print(f.read())

# import files from the user filesystem
sys.path.append("/userfs")
import usermod1

# unmount and undo path addition
uos.umount("/userfs")
sys.path.pop()



try:
    import ujson as json
except ImportError:
    try:
        import json
    except ImportError:
        print("SKIP")
        raise SystemExit

print(json.dumps(1.2))
print(json.dumps({1.5: "hi"}))



# test _thread lock objects with multiple threads
#
# SPDX-FileCopyrightText: Copyright (c) 2016 Damien P. George on behalf of Pycom Ltd
#
# SPDX-License-Identifier: MIT

try:
    import utime as time
except ImportError:
    import time
import _thread

lock = _thread.allocate_lock()


def thread_entry():
    lock.acquire()
    print("have it")
    lock.release()


# spawn the threads
for i in range(4):
    _thread.start_new_thread(thread_entry, ())

# wait for threads to finish
time.sleep(1)
print("done")



@micropython.asm_thumb
def lsl1(r0):
    lsl(r0, r0, 1)


print(hex(lsl1(0x123)))


@micropython.asm_thumb
def lsl23(r0):
    lsl(r0, r0, 23)


print(hex(lsl23(1)))


@micropython.asm_thumb
def lsr1(r0):
    lsr(r0, r0, 1)


print(hex(lsr1(0x123)))


@micropython.asm_thumb
def lsr31(r0):
    lsr(r0, r0, 31)


print(hex(lsr31(0x80000000)))


@micropython.asm_thumb
def asr1(r0):
    asr(r0, r0, 1)


print(hex(asr1(0x123)))


@micropython.asm_thumb
def asr31(r0):
    asr(r0, r0, 31)


print(hex(asr31(0x80000000)))



# test setting the thread stack size
#
# SPDX-FileCopyrightText: Copyright (c) 2016 Damien P. George on behalf of Pycom Ltd
#
# SPDX-License-Identifier: MIT

import sys
import _thread

# different implementations have different minimum sizes
if sys.implementation.name == "micropython":
    sz = 2 * 1024
else:
    sz = 512 * 1024


def foo():
    pass


def thread_entry():
    foo()
    with lock:
        global n_finished
        n_finished += 1


# reset stack size to default
_thread.stack_size()

# test set/get of stack size
print(_thread.stack_size())
print(_thread.stack_size(sz))
print(_thread.stack_size() == sz)
print(_thread.stack_size())

lock = _thread.allocate_lock()
n_thread = 2
n_finished = 0

# set stack size and spawn a few threads
_thread.stack_size(sz)
for i in range(n_thread):
    while True:
        try:
            _thread.start_new_thread(thread_entry, ())
            break
        except OSError:
            pass

# reset stack size to default (for subsequent scripts on baremetal)
_thread.stack_size()

# busy wait for threads to finish
while n_finished < n_thread:
    pass
print("done")



# test try-else statement

# base case
try:
    print(1)
except:
    print(2)
else:
    print(3)

# basic case that should skip else
try:
    print(1)
    raise Exception
except:
    print(2)
else:
    print(3)

# uncaught exception should skip else
try:
    try:
        print(1)
        raise ValueError
    except TypeError:
        print(2)
    else:
        print(3)
except:
    print('caught')

# nested within outer try
try:
    print(1)
    try:
        print(2)
        raise Exception
    except:
        print(3)
    else:
        print(4)
except:
    print(5)
else:
    print(6)

# nested within outer except, one else should be skipped
try:
    print(1)
    raise Exception
except:
    print(2)
    try:
        print(3)
    except:
        print(4)
    else:
        print(5)
else:
    print(6)

# nested within outer except, both else should be skipped
try:
    print(1)
    raise Exception
except:
    print(2)
    try:
        print(3)
        raise Exception
    except:
        print(4)
    else:
        print(5)
else:
    print(6)



# test implicit scoping rules

# implicit nonlocal, with variable defined after closure
def f():
    def g():
        return x # implicit nonlocal
    x = 3 # variable defined after function that closes over it
    return g
print(f()())

# implicit nonlocal at inner level, with variable defined after closure
def f():
    def g():
        def h():
            return x # implicit nonlocal
        return h
    x = 4 # variable defined after function that closes over it
    return g
print(f()()())

# local variable which should not be implicitly made nonlocal
def f():
    x = 0
    def g():
        x # local because next statement assigns to it
        x = 1
    g()
try:
    f()
except NameError:
    print('NameError')



"""
categories: Types,bytes
description: bytes objects support .format() method
cause: MicroPython strives to be a more regular implementation, so if both `str` and `bytes` support ``__mod__()`` (the % operator), it makes sense to support ``format()`` for both too. Support for ``__mod__`` can also be compiled out, which leaves only ``format()`` for bytes formatting.
workaround: If you are interested in CPython compatibility, don't use ``.format()`` on bytes objects.
"""
print(b"{}".format(1))



# test out-of-memory with malloc
l = list(range(1000))
try:
    1000000000 * l
except MemoryError:
    print('MemoryError')
print(len(l), l[0], l[-1])

# test out-of-memory with realloc
try:
    [].extend(range(1000000000))
except MemoryError:
    print('MemoryError')



# Doing some operation on bytearray
# Inplace - the most memory efficient way
import bench


def test(num):
    for i in iter(range(num // 10000)):
        ba = bytearray(b"\0" * 1000)
        for i in range(len(ba)):
            ba[i] += 1


bench.run(test)
